/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.data;

import amberstone.lang.BytesFunction;

import amberstone.lang.Nullable;

import java.io.Externalizable;
import java.io.ObjectOutput;
import java.io.ObjectInput;
import java.io.IOException;

public class ByteArray implements Externalizable,
			                      BytesFunction,
			                      Nullable
{
	public boolean clone;
	private byte[] buffer;
	public int offset;
	private int pointer;
	public int length;

	private int hash;
	private int intValue;
	private long longValue;
	private boolean hashComputed;
	private boolean intComputed;
	private boolean longComputed;

	public ByteArray()
	{
	}

	public ByteArray(int length)
	{
		this.buffer = new byte[length];
	}

	public ByteArray(String s)
	{
		buffer = s.getBytes();
		length = buffer.length;
	}

	public ByteArray(byte[] buf)
	{
		this.buffer = buf;
		this.length = buf.length;
	}

	public ByteArray getBytes()
	{
		return this;
	}

	public void setHash(int hash)
	{
		this.hash = hash;
		this.hashComputed = true;
	}

	public boolean isNull()
	{
		return this.length == 0;
	}

	public void set(byte[] buffer, int offset, int length)
	{
		this.buffer = buffer;
		this.offset = this.pointer = offset;
		this.length = length;
		this.hash = 0;
		this.hashComputed = this.intComputed = this.longComputed = false;
	}

	public void set(ByteArray bytes)
	{
		this.buffer   = bytes.buffer;
		this.offset   = bytes.offset;
		this.pointer  = bytes.pointer;
		this.length   = bytes.length;
		this.hash = 0;
		this.hashComputed = this.intComputed = this.longComputed = false;
	}

	public int getOffset()
	{
		return this.offset;
	}

	public int getPointer()
	{
		return this.pointer;
	}

	/**
	*  Sets the fixed offset of the ByteArray and the internal pointer to a specific index.
	**/

	public void seek(int x)
	{
		this.pointer = this.offset = x;
	}


	/**
	* Moves the internal offset and pointer x number of bytes
	**/

	public int move(int x)
	{
		this.pointer = this.offset+=x;
		return this.offset;
	}

	/**
	* Sets the internal pointer x number of bytes from the offset.
	**/

	public void point(int x)
	{
		this.pointer = this.offset+x;
	}


	/**
	* Moves the internal pointer x number of bytes. Does not change the offset.
	**/
	public void skip(int x)
	{
		this.pointer += x;
	}

	public int getLength()
	{
		return this.length;
	}

	public byte[] getBuffer()
	{
		return this.buffer;
	}

	public byte[] toByteArray()
	{
		byte[] b = new byte[length];
		System.arraycopy(buffer, offset, b, 0, length);
		return b;
	}

	public void reset()
	{
		this.pointer = this.offset;
	}

	public void clear()
	{
		this.pointer = this.offset = this.length = this.hash = 0;
		this.hashComputed = this.intComputed = this.longComputed = false;
	}

	public void setLength(int length)
	{
		this.length = length;
	}

	public void setBuffer(byte[] buffer)
	{
		this.buffer = buffer;
	}

	public int indexOf(ByteArray findBytes)
	{
		return indexOf(findBytes, 0);
	}

	public static void main(String args[])
	{
		/*
		ByteArray source = new ByteArray(args[0]);
		ByteArray target = new ByteArray(args[1]);

		int index = source.indexOf(target);
		//System.out.println("index is:"+index);
		*/

		ByteArray b = new ByteArray(args[0]);
		System.out.println("test:"+new Long(b.parseLong()).toString());
	}

	public void trim()
	{
		ltrim();
		rtrim();
	}

	public void ltrim()
	{
		int count=offset+length;
		for(int i=offset; i<count; i++)
		{
			if(buffer[i] < 33)
			{
				this.offset++;
				this.length--;
			}
			else
			{
				break;
			}
		}

		this.pointer = this.offset;
	}

	public void rtrim()
	{
		for(int i=(offset+this.length-1); i>=offset; i--)
		{
			if(buffer[i] < 33)
			{
				this.length--;
			}
			else
			{
				break;
			}
		}
	}

    public int indexOf(ByteArray findBytes,
                       int fromIndex)
    {
		int start = offset+fromIndex;

		byte[] findBuffer = findBytes.getBuffer();
		int findOffset    = findBytes.offset;
		int findLength    = findBytes.length;

        byte firstByte = findBuffer[findOffset];

        int dif = (length-(fromIndex)) - findLength;
        if(dif < 0)
        {
			return -1;
		}

        int max = start + dif;
		boolean match = true;
		int findStart = findOffset+1;
		int findEnd = findOffset+findLength;

        for(int i = start; i <= max; i++)
        {
			if(buffer[i] == firstByte)
			{
				int c = i+1;
				match = true;
            	for(int j=findStart; j<findEnd; j++, c++)
            	{
					if(buffer[c] != findBuffer[j])
					{
						match = false;
						break;
					}
				}

				if(match) return i-offset;
			}
        }

        return -1;
    }

	public void writeExternal(ObjectOutput out)
		throws IOException
	{
		out.writeInt(this.length);
		out.write(buffer, offset, length);
	}

	public void readExternal(ObjectInput in)
		throws IOException, ClassNotFoundException
	{
		this.length = in.readInt();
		buffer = new byte[length];
		in.read(buffer, 0, length);
	}

	public int hashCode()
	{
		if(hashComputed) return this.hash;
		int h = 0;
		int index = (offset+length);
		while(index>offset)
		{
			h = (31*h) + buffer[--index];
		}

		hash = h;
		hashComputed = true;

		return h;
	}

	public boolean equals(Object o)
	{
		ByteArray compareBytes = (ByteArray)o;

		if(this.length == compareBytes.length)
		{
			int index = (offset+length);
			int compareIndex = (compareBytes.offset+compareBytes.length);

			while(index>offset)
			{
				if(buffer[--index] != compareBytes.buffer[--compareIndex])
				{
					return false;
				}
			}

			return true;
		}
		else
		{
			return false;
		}
	}

	public void writeBytes(ByteArray data)
	{
		byte[] dataBuffer = data.buffer;
		int dataPointer   = data.pointer;
		int dataLength    = data.length;
		int i=0;
		for(i=0; (i<dataLength && i<buffer.length); i++)
		{
			buffer[pointer++] = dataBuffer[dataPointer++];
		}

		this.length += i;
	}

	public void writeBytes(ByteArray data, int maxLength)
	{
		byte[] dataBuffer = data.buffer;
		int dataPointer   = data.pointer;
		int dataLength    = data.length;
		int i=0;

		for(i=0; i<dataLength; i++)
		{
			buffer[pointer++] = dataBuffer[dataPointer++];
		}

		this.length+=dataLength;
	}

	public void write(byte[] data, int offset, int length)
	{
		for(int i=0; i<length; ++i)
		{
			buffer[pointer++] = data[offset++];
		}
		this.length += length;
	}

	public void pad(int num)
	{
		for(int i=0; i<num; i++)
		{
			buffer[pointer++] = 0;
		}
	}

	public int readInt()
	{
		return ((buffer[pointer++] & 0xff) << 24) + ((buffer[pointer++] & 0xff) << 16) + ((buffer[pointer++] & 0xff) << 8) + (buffer[pointer++] & 0xff);
	}

	public int readInt(int point)
	{
		pointer = this.offset+point;
		return ((buffer[pointer++] & 0xff) << 24) + ((buffer[pointer++] & 0xff) << 16) + ((buffer[pointer++] & 0xff) << 8) + (buffer[pointer] & 0xff);
	}

	public void writeInt(int i)
	{
		buffer[pointer++] = (byte)((i >>> 24) & 0xFF);
		buffer[pointer++] = (byte)((i >>> 16) & 0xFF);
		buffer[pointer++] = (byte)((i >>>  8) & 0xFF);
		buffer[pointer++] = (byte)((i >>>  0) & 0xFF);
		this.length += 4;
	}

	public void writeInt3(int i)
	{
		buffer[pointer++] = (byte)((i >>> 16) & 0xFF);
		buffer[pointer++] = (byte)((i >>>  8) & 0xFF);
		buffer[pointer++] = (byte)((i >>>  0) & 0xFF);
		this.length += 3;
	}

	public int readInt3(int point)
	{
		pointer = this.offset+point;
		return (int)(((buffer[pointer++] & 0xff) << 16) + ((buffer[pointer++] & 0xff) << 8) + (buffer[pointer] & 0xff));
	}

	public int readInt3()
	{
		return (int)(((buffer[pointer++] & 0xff) << 16) + ((buffer[pointer++] & 0xff) << 8) + (buffer[pointer++] & 0xff));
	}

	public void writeInt2(int i)
	{
		buffer[pointer++] = (byte)((i >>>  8) & 0xFF);
		buffer[pointer++] = (byte)((i >>>  0) & 0xFF);
		this.length += 2;
	}

	public int readInt2(int point)
	{
		pointer = this.offset+point;
		return (int)(((buffer[pointer++] & 0xff) << 8) + (buffer[pointer] & 0xff));
	}

	public int readInt2()
	{
		return (int)(((buffer[pointer++] & 0xff) << 8) + (buffer[pointer++] & 0xff));
	}

	public void writeInt1(int i)
	{
		buffer[pointer++] = (byte)((i >>>  0) & 0xFF);
		this.length++;
	}

	public int readInt1(int point)
	{
		pointer = this.offset+point;
		return ((int)(buffer[pointer] & 0xff));
	}

	public int readInt1()
	{
		return ((int)(buffer[pointer++] & 0xff));
	}

	public int shift1()
	{
		int shift = ((int)(buffer[offset++] & 0xff));
		pointer++;
		return shift;
	}

	public int read()
	{
		return buffer[pointer++];
	}

	public void writeByte(byte b)
	{
		buffer[pointer++] = b;
		this.length++;
	}

	public void writeLong(long l)
	{
		buffer[pointer++] = (byte)(l >>> 56);
		buffer[pointer++] = (byte)(l >>> 48);
		buffer[pointer++] = (byte)(l >>> 40);
		buffer[pointer++] = (byte)(l >>> 32);
		buffer[pointer++] = (byte)(l >>> 24);
		buffer[pointer++] = (byte)(l >>> 16);
		buffer[pointer++] = (byte)(l >>>  8);
		buffer[pointer++] = (byte)(l >>>  0);
		this.length += 8;
	}

	public long readLong(int point)
	{
		pointer = this.offset+point;
		return (((long)buffer[pointer++] << 56) + ((long)(buffer[pointer++] & 255) << 48) + ((long)(buffer[pointer++] & 255) << 40) +
		        ((long)(buffer[pointer++] & 255) << 32) + ((long)(buffer[pointer++] & 255) << 24) + ((buffer[pointer++] & 255) << 16) +
		        ((buffer[pointer++] & 255) <<  8) + (buffer[pointer++] & 255));

	}

	public long readLong()
	{
		return (((long)buffer[pointer++] << 56) + ((long)(buffer[pointer++] & 255) << 48) + ((long)(buffer[pointer++] & 255) << 40) +
		        ((long)(buffer[pointer++] & 255) << 32) + ((long)(buffer[pointer++] & 255) << 24) + ((buffer[pointer++] & 255) << 16) +
		        ((buffer[pointer++] & 255) <<  8) + (buffer[pointer++] & 255));

	}

	public long parseLong() throws NumberFormatException
	{
		if(this.longComputed) return this.longValue;

		long power = 1L;
		long num = 0L;
		boolean negative = false;

		for(int i=(offset+this.length-1); i>=offset; i--)
		{
			int b = buffer[i];
			if(b > 47 && b < 58)
			{
				num = num+((b-48)*power);
			}
			else if(b == 45)
			{
				negative = true;
			}
			else if(b == 46)
			{
				power = 1L;
				num = 0L;
			}
			else
			{
				byte[] bad = {(byte)b};
				throw new NumberFormatException("NumberFormatException:"+new String(bad));
			}
			power = power*10;
		}

		if(negative)
			num = -num;

		this.longValue = num;
		this.longComputed = true;
		return num;
	}

	public int parseInt() throws NumberFormatException
	{
		if(this.intComputed) return this.intValue;

		int power = 1;
		int num = 0;
		boolean negative = false;

		for(int i=(offset+this.length-1); i>=offset; i--)
		{
			int b = buffer[i];
			if(b > 47 && b < 58)
			{
				num = num+((b-48)*power);
			}
			else if(b == 45)
			{
				negative = true;
			}
			else
			{
				byte[] bad = {(byte)b};
				throw new NumberFormatException("NumberFormatException:"+new String(bad));
			}

			power = power*10;
		}

		if(negative)
			num = -num;

		this.intValue = num;
		this.intComputed = true;
		return num;
	}

	public void toLowerCase()
	{
		byte add = (byte)32;
		int c=0;
		for(int i=offset; c<this.length; ++i, ++c)
		{
			byte b = buffer[i];

			if(b > 64 && b < 91)
			{
				buffer[i] = (byte) (b + add);
			}
		}
	}

	public ByteArray cloneBytes()
	{
		ByteArray clone = new ByteArray(length);
		clone.writeBytes(this);
		clone.reset();
		clone.hashComputed = hashComputed;
		clone.hash = hash;
		clone.clone = true;
		return clone;
	}

	public String toString()
	{
		return new String(buffer, offset, length);
	}
}