/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/
package amberstone.data;

import amberstone.lang.IntFunction;
import amberstone.data.fields.Bytes;

import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.zip.DeflaterOutputStream;
import java.util.zip.InflaterInputStream;
import java.util.zip.Deflater;
import java.util.zip.Inflater;

import java.io.*;

public class DataRecord
{
	public Exception e;
	private File dir;
	private Field[] fields;
	private Field[] activeFields;
	private Map<String, Field>fieldMap = new HashMap();
	private int readCount;
	private int writeCount;
	private int maxLength;
	private int minLength;

	private int blockRecords;
	private byte[] readBytes;

	public DataRecord(List<Field> fieldList)
	{
		this.fields = new Field[fieldList.size()];

		for(int i=0; i<fields.length; i++)
		{
			Field field = fieldList.get(i);
			this.maxLength += field.maxLength();
			this.minLength += field.minLength();
			fields[i] = field;
			fieldMap.put(field.getName(), field);
		}
	}

	public DataRecord()
	{

	}

	public void getData(DataRecord transfer)
	{
		this.dir = transfer.dir;
		this.blockRecords = transfer.blockRecords;
		this.readCount = 0;
		for(int i=0; i<activeFields.length; i++)
		{
			Field rfield = activeFields[i];
			Field tfield = transfer.getField(rfield.getName());
			ByteArray tbytes = tfield.getBuffer();
			byte[] buffer = tbytes.getBuffer();
			ByteArray rbytes = rfield.getBuffer();
			rbytes.set(buffer, tbytes.getOffset(), tbytes.getLength());
		}
	}

	public DataRecord cloneRecord()
	{
		List<Field> clonedFields = new ArrayList();
		for(int i=0; i<fields.length;i++)
		{
			Field cfield = fields[i].cloneField();
			clonedFields.add(cfield);

		}

		DataRecord dr =  new DataRecord(clonedFields);
		return dr;
	}

	public void setDir(File dir)
	{
		this.dir = dir;
	}

	public File getDir()
	{
		return this.dir;
	}

	public void writeBlock() throws IOException
	{
		if(writeCount == 0) return;

		PrintWriter stats = null;

		try
		{
			File sfile = new File(dir, "stats");
			stats = new PrintWriter(new FileWriter(sfile));
			stats.println(this.writeCount);
			this.writeCount = 0;
		}
		finally
		{
			stats.close();
		}

		for(int i=0; i<fields.length; i++)
		{
			DeflaterOutputStream out = null;

			try
			{
				Field f = fields[i];
				ByteArray bytes = f.getBuffer();
				String name = f.signiture();
				File file = new File(dir, name+".dat");

				Deflater deflater = null;
				if(f instanceof Bytes)
				{
					deflater = new Deflater(Deflater.BEST_SPEED, true);
				}
				else
				{
					deflater = new Deflater(Deflater.HUFFMAN_ONLY, true);
				}

				out = new  DeflaterOutputStream(new FileOutputStream(file), deflater, (int)bytes.getLength()/2);
				out.write(bytes.getBuffer(), 0, bytes.getLength());
			}
			finally
			{
				out.close();
			}
		}
	}

	public int readBlock() throws IOException
	{
		BufferedReader reader = null;
		try
		{
			File f = new File(dir, "stats");
			if(f.exists())
			{
				reader = new BufferedReader(new FileReader(f));
				String l = reader.readLine();
				blockRecords = Integer.parseInt(l);
			}
			else
			{
				blockRecords = 0;
			}
		}
		finally
		{
			reader.close();
		}

		int offset = 0;
		for(int i=0; i<activeFields.length; i++)
		{
			InflaterInputStream in = null;

			try
			{
				Field f = activeFields[i];
				ByteArray buf = f.getBuffer();
				String name = f.signiture();
				File file = new File(dir, name+".dat");
				if(file.exists())
				{
					//long begin = System.nanoTime();

					int read = 0;
					in = new InflaterInputStream(new FileInputStream(file), new Inflater(true), (int)file.length());
					int bytesToRead = readBytes.length-offset;
					int totalBytes = 0;
					int initialOffset = offset;
					while((read = in.read(readBytes, offset, bytesToRead)) != -1)
					{
						offset+=read;
						totalBytes+=read;
						bytesToRead = readBytes.length-offset;
					}

					f.initBuffer(readBytes, initialOffset, totalBytes);

					//long end = System.nanoTime();
					//System.out.println("Field Read Time: "+Double.toString((double)(end-begin)/(double)1000000));

				}
				else
				{
					f.initBuffer(readBytes, 0, 0);
				}
			}
			finally
			{
				if(in != null)
				{
					in.close();
				}
			}
		}

		readCount = 0;

		return blockRecords;
	}

	public int initWrite(int blockSize)
	{
		int recordsPerBlock = 0;
		if(this.maxLength == this.minLength)
		{
			recordsPerBlock = (int)(blockSize/this.maxLength);
		}
		else
		{
			recordsPerBlock = (int)((blockSize*2)/this.maxLength);
		}

		for(int i=0; i<fields.length; i++)
		{
			int m = fields[i].maxLength();
			byte[] bs = new byte[m*recordsPerBlock];
			fields[i].initBuffer(bs, 0, 0);
		}

		return recordsPerBlock;
	}

	public void initRead(int blockSize)
	{
		this.readBytes = new byte[blockSize];
	}

	public byte[] getReadBytes()
	{
		return readBytes;
	}

	public void reset()
	{
		for(int i=0; i<fields.length; i++)
		{
			fields[i].reset();
		}
	}

	public Field[] getActiveFields()
	{
		return this.activeFields;
	}

	public Field[] getFields()
	{
		return this.fields;
	}

	public Field getField(String name)
	{
		return this.fieldMap.get(name);
	}

	public Field getField(int index)
	{
		return this.fields[index];
	}

	public int maxLength()
	{
		return this.maxLength;
	}

	public int minLength()
	{
		return this.minLength;
	}

	public int fieldCount()
	{
		return this.fields.length;
	}

	public boolean readRecord()
	{
		if(readCount < blockRecords)
		{
			for(int i=0; i<activeFields.length; i++)
			{
				activeFields[i].readField();
			}
			++readCount;

			return true;
		}
		else
		{
			return false;
		}
	}

	public int prepareRecord() throws Exception
	{
		int c = 0;

		for(int i=0; i<fields.length; i++)
		{
			Field field = fields[i];
			c += field.prepareWrite();
		}

		if(c < minLength)
		{
			return -1;
		}
		else
		{
			return c;
		}
	}

	public int writeRecord() throws Exception
	{
		int c = 0;
		for(int i=0; i<fields.length; i++)
		{
			Field field = fields[i];
			c += field.writeField();
		}

		++writeCount;
		return c;
	}

	public void setActiveFields(Field[] activeFields)
	{
		this.activeFields = activeFields;
	}
}




