/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.dictionary;

import jdbm.RecordManager;
import jdbm.RecordManagerFactory;
import jdbm.helper.FastIterator;
import jdbm.htree.HTree;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Properties;

import amberstone.data.ByteArray;
import amberstone.resources.ResourceIterator;
import amberstone.resources.DiskResourceIterator;
import amberstone.util.MutableInteger;
import amberstone.util.LRUCache;

public class DiskDictionary implements Dictionary
{
	private ByteArray sequenceKey = new ByteArray("_SEQ");
	private Sequence sequence;
	private RecordManager recman;
	private HTree terms;
	private HTree ids;
	private int puts;
	private MutableInteger ikey = new MutableInteger();
	private ByteArray notFound = new ByteArray("Key Not Found");

	public static String dir;
	public static int commitEvery;
	public static boolean transactions;

	private String name;

	private Properties props = new Properties();

	public DiskDictionary(String name)
		throws IOException
	{
		this.name = name;
		props.put("jdbm.cache.size", "5000");
		String t = transactions ? "FALSE" : "TRUE";
		props.put("jdbm.disableTransactions", t);
	}

	public Dictionary getDictionary()
	{
		return this;
	}

	public ResourceIterator terms() throws IOException
	{
		return new DiskResourceIterator(terms.keys());
	}

	public int termCount()
	{
		return sequence.getValue();
	}

	public String getName()
	{
		return this.name;
	}

	public void open()
				throws IOException
	{
		recman = RecordManagerFactory.createRecordManager(this.dir+File.separatorChar+this.name, this.props);

		//Load the terms table
		long recid = recman.getNamedObject("terms");
		if (recid != 0)
		{
			this.terms = HTree.load(recman, recid);
		}
		else
		{
			terms = HTree.createInstance(recman);
		    recman.setNamedObject("terms", terms.getRecid());
        }

		sequence = (Sequence)terms.get(sequenceKey);

		if(sequence == null)
		{
			sequence = new Sequence();
		}

        //load the id table
		recid = recman.getNamedObject("ids");
		if (recid != 0)
		{
			ids = HTree.load(recman, recid);
		}
		else
		{
			ids = HTree.createInstance(recman);
		    recman.setNamedObject("ids", ids.getRecid());
        }
	}

	public void drop() throws IOException
	{
		File db = new File(this.dir+File.separatorChar+this.name+".db");
		File log = new File(this.dir+File.separatorChar+this.name+".lg");
		db.delete();
		log.delete();
	}

	public void close() throws IOException
	{
		commit();
		this.recman.close();
	}

	public void commit() throws IOException
	{
		this.terms.put(sequenceKey, sequence);
		this.recman.commit();
		this.puts = 0;
	}

	public int putTerm(ByteArray term) throws Exception
	{

		MutableInteger id = (MutableInteger)terms.get(term);

		if(id == null)
		{
			if(!term.clone)
			{
				term = term.cloneBytes();
			}

			int idNum = sequence.nextId();
			id = new MutableInteger(idNum);
			terms.put(term, id);
			ids.put(id, term);

			++puts;
			if(puts > commitEvery)
			{
				commit();
			}
		}

		return id.getInt();
	}

	public ByteArray getTerm(int id) throws Exception
	{
		ikey.setInt(id);
		ByteArray ba = (ByteArray)ids.get(ikey);
		if(ba == null)
		{
			return notFound;
		}
		else
		{
			return ba;
		}
	}

	public int getId(ByteArray term) throws IOException
	{
		MutableInteger i = (MutableInteger)terms.get(term);
		if(i != null)
		{
			return i.getInt();
		}
		else
		{
			return 0;
		}
	}
}

