/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.file;

import java.io.File;
import java.io.FileFilter;
import java.util.Comparator;
import java.util.ArrayList;
import java.util.Arrays;

public class FileHandler
{
	protected Comparator fileStatSorter = new FileStatSorter();
	protected Comparator dirSorter = new DirSorter();

	public FileStat[] getFiles(File root, FileFilter fileFilter, boolean recurse)
	{
		ArrayList<FileStat> allFiles = new ArrayList();
		descend(root, allFiles, fileFilter, recurse);
		FileStat[] stats = allFiles.toArray(new FileStat[allFiles.size()]);
		Arrays.sort(stats, fileStatSorter);
		return stats;
	}

	public File[] getDirs(File root)
	{
		ArrayList<File> dirs = new ArrayList();
		File[] files = root.listFiles();
		for(int i=0;i<files.length; i++)
		{
			if(files[i].isDirectory())
			{
				dirs.add(files[i]);
			}
		}

		File[] sortedDirs = dirs.toArray(new File[dirs.size()]);
		Arrays.sort(sortedDirs, dirSorter);

		return sortedDirs;
	}

	private void descend(File file, ArrayList<FileStat> allFiles, FileFilter fileFilter, boolean recurse)
	{
		File[] files = file.listFiles(fileFilter);
		for(int i=0;i<files.length; i++)
		{
			File f = files[i];
			if(f.isDirectory() && recurse)
			{
				descend(f, allFiles, fileFilter, recurse);
			}
			else if(f.isFile())
			{
				FileStat fstat = new FileStat();
				fstat.fileNumber = i;
				fstat.fileLength = (int)f.length();
				String fname = f.getName();
				fstat.path = f.getParentFile().getAbsolutePath();
				fstat.fullName = fname;
				String nameParts[] = fname.split("\\.");

				int partsIndex = 0;
				fstat.name = Integer.parseInt(nameParts[partsIndex++]);
				if(nameParts.length > 2)
				{
					fstat.partNumber = Integer.parseInt(nameParts[partsIndex++]);
				}
				else
				{
					fstat.partNumber = 1;
				}

				if(nameParts.length > 3)
				{
					fstat.numCols = Integer.parseInt(nameParts[partsIndex++]);
					fstat.recLength = Integer.parseInt(nameParts[partsIndex++]);
					fstat.time = Integer.parseInt(nameParts[partsIndex++]);
				}
				fstat.extension = nameParts[partsIndex];

				fstat.file = f;
				allFiles.add(fstat);
			}
		}
	}

	class FileStatSorter implements Comparator
	{
		public int compare(Object o1, Object o2)
		{
			FileStat f1 = (FileStat)o1;
			FileStat f2 = (FileStat)o2;
			int i1 = f1.name;
			int i2 = f2.name;

			if(i1 == i2)
			{
				if(f1.partNumber < f2.partNumber)
				{
					return -1;
				}
				else
				{
					return 1;
				}
			}
			else if(i1 < i2)
			{
				return -1;
			}
			else
			{
				return 1;
			}
		}
	}

	class DirSorter implements Comparator
	{
		public int compare(Object o1, Object o2)
		{
			File dir1 = (File)o1;
			File dir2 = (File)o2;

			int num1 = Integer.parseInt(dir1.getName());
			int num2 = Integer.parseInt(dir2.getName());

			if(num1 < num2)
			{
				return -1;
			}
			else
			{
				return 1;
			}
		}
	}
}