/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.lang;

import amberstone.system.Logger;

import java.util.Stack;
import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Properties;
import java.io.StreamTokenizer;
import java.io.StringReader;
import java.math.BigDecimal;

public class Compiler
{
	private Properties macros;
	private Properties lang;
	private Logger logger;
	private int logLevel;


	public void setMacros(Properties macros)
	{
		this.macros = macros;
	}

	public void setLang(Properties lang)
	{
		this.lang = lang;
	}

	public void setLogger(int logLevel, Logger logger)
	{
		this.logLevel = logLevel;
		this.logger = logger;
	}

	private StreamTokenizer getTokenizer(String statement)
	{
		StreamTokenizer tokens = new StreamTokenizer(new StringReader(statement));
		tokens.wordChars(65, 90);   // A-Z
		tokens.wordChars(97, 122); // a-z
		tokens.wordChars(48, 57);  // 1-9
		tokens.wordChars(43, 43);  // +
		tokens.wordChars('-', '-');  //
		tokens.wordChars('#', '#');  //
		tokens.wordChars('@', '@');  //
		tokens.wordChars(36, 36);  // $
		tokens.wordChars(95, 95);    //_
		tokens.quoteChar(39);   //'
		tokens.ordinaryChars(28,29); // ()
		tokens.ordinaryChars(44,44); // ,
		tokens.whitespaceChars(32, 32); // space

		return tokens;
	}

	private List<String> tokenList(String statement) throws Exception
	{
		List<String> tokenList = new ArrayList();

		int parens = 0;
		int quotes = 0;
		StreamTokenizer tokens = getTokenizer(statement);

		int type = -1;
		while((type = tokens.nextToken()) != tokens.TT_EOF)
		{
			if(type == tokens.TT_WORD)
			{
				tokenList.add(tokens.sval);
			}
			else if(type == 39)
			{
				tokenList.add("'"+tokens.sval+"'");
			}
			else if(type == tokens.TT_NUMBER)
			{
				tokenList.add(Double.toString(tokens.nval));
			}
			else if(type== '(')
			{
				parens++;
				tokenList.add("(");
			}
			else if(type == ')')
			{
				parens--;
				tokenList.add(")");
			}
			else if(type == ',')
			{
				tokenList.add(", ");
			}
			else
			{
				byte[] b = {(byte)type};
				throw new Exception("Invalid token:"+new String(b)+" : "+type);
			}
		}

		if(parens != 0)
		{
			throw new UnbalancedParensException(parens);
		}

		return tokenList;
	}

	private void checkQuotes(String statement) throws Exception
	{
		int quotes = 0;
		for(int i=0; i<statement.length(); i++)
		{
			if(statement.charAt(i) == '\'')
			{
				quotes++;
			}
		}

		if(quotes % 2 != 0)
		{
			throw new Exception("Unbalanced quotes");
		}
	}


	public Object compile(String statement) throws CompilationException

	{
		int tokenCount = 0;
		List<String> tokenList = null;
		Stack stack = new Stack();

		try
		{

			StreamTokenizer tokens = getTokenizer(statement);
			int type = -1;
			checkQuotes(statement);
			tokenList = tokenList(statement);

			while((type = tokens.nextToken()) != tokens.TT_EOF)
			{
				if(type == tokens.TT_WORD)
				{
					if(tokens.sval.equals("true") || tokens.sval.equals("false"))
					{
						BooleanLiteral bool = new BooleanLiteral(Boolean.parseBoolean(tokens.sval));
						stack.push(bool);
					}
					else if(tokens.sval.startsWith("$"))
					{
						Col col = new Col();
						StringLiteral s = new StringLiteral(tokens.sval.substring(1));
						s.setTokenIndex(tokenCount);
						Function[] f = {s};
						col.params(f);
						stack.push(col);
					}
					else if(tokens.sval.startsWith("#"))
					{
						DataColumn col = new DataColumn();
						System.out.println(tokens.sval.substring(1));
						NumberLiteral s = new NumberLiteral(Integer.parseInt(tokens.sval.substring(1)));
						Function[] f = {s};
						col.params(f);
						stack.push(col);
					}
					else if(tokens.sval.startsWith("@"))
					{
						TransactionFunction col = new TransactionFunction();
						StringLiteral s = new StringLiteral(tokens.sval.substring(1));
						Function[] f = {s};
						col.params(f);
						stack.push(col);
					}
					else
					{
						String className = lang.getProperty(tokens.sval);

						if(className == null)
						{
							throw new Exception("The function '"+tokens.sval+"' is undefined.");
						}

						Function function = (Function)Class.forName(className).newInstance();
						function.setCallName(tokens.sval);
						function.setTokenIndex(tokenCount);
						stack.push(function);
					}
					tokenCount++;

				}
				else if(type == 39)
				{
					StringLiteral strlit = new StringLiteral(tokens.sval);
					strlit.setTokenIndex(tokenCount);
					tokenCount++;
					stack.push(strlit);
				}
				else if(type == tokens.TT_NUMBER)
				{
					NumberLiteral num = new NumberLiteral(tokens.nval);
				    num.setTokenIndex(tokenCount);
				    tokenCount++;
					stack.push(num);
				}
				else if(type== '(')
				{
					tokenCount++;
					stack.push("(");
				}
				else if(type == ')')
				{
					reduce(stack);
					tokenCount++;
				}
				else if(type == ',')
				{
					tokenCount++;
				}
			}
		}
		catch(CompilationException ce)
		{
			ce.setStatement(statement);
			ce.setTokens(tokenList);
			throw ce;
		}
		catch(Exception e)
		{
			CompilationException c = new CompilationException(tokenCount, e);
			c.setTokens(tokenList);
			c.setStatement(statement);
			throw c;
		}

		return stack.pop();
	}

	public void reduce(Stack stack) throws Exception
	{
		LinkedList parameters = new LinkedList();
		while(true)
		{
			Object o = stack.pop();

			if(o instanceof String && o.equals("("))
			{
				Function c = (Function)stack.peek();
				c.link(parameters.toArray());

				if(c instanceof MacroPointer)
				{
					stack.pop();
					String macroName = c.toString();
					String macro = macros.getProperty(macroName);

					if(macro == null)
					{
						throw new CompilationException(c.getTokenIndex()+2, "The macro '"+macroName+"' is undefined.");
					}

					//logger.log("compiling macro:"+macroName);
					try
					{
						stack.push(compile(macro));
					}
					catch(Exception e)
					{
						throw new Exception("Compilation error in macro '"+macroName+"'", e);
					}
				}

				break;
			}
			else
			{
				parameters.addFirst(o);
			}
		}
	}
}