/*
*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.lang;

import amberstone.data.ByteArray;
import java.math.BigDecimal;

public class Scaler extends Function implements IntFunction, LongFunction
{
	private BytesFunction bytesFunc;
	private int scale;
	private boolean round;

	private static final FunctionType[][] allow = { { FunctionType.BYTES_FUNCTION, FunctionType.NUMBER_LITERAL },
													{ FunctionType.BYTES_FUNCTION, FunctionType.NUMBER_LITERAL, FunctionType.BOOLEAN_LITERAL }};

	public FunctionType[][] strict()
	{
		return allow;
	}

	public void params(Function[] params) throws ParameterException
	{
		this.bytesFunc = (BytesFunction)params[0];
		this.scale = ((NumberLiteral)params[1]).getInt();
		if(params.length == 3)
		{
			round = ((BooleanLiteral)params[2]).getBoolean();
		}
		else
		{
			round = true;
		}
	}

	public int getScale()
	{
		return scale;
	}

	public int getInt() throws FunctionException
	{
		ByteArray bytes = bytesFunc.getBytes();
		int offset = bytes.getOffset();
		int numBytes = bytes.getLength();
		byte[] buffer = bytes.getBuffer();
		boolean negative = false;
		boolean roundUp = false;
		int pointPosition = 0;
		int num = 0;

		int startPoint = pointPosition = offset+numBytes-1;
		//Scan and find the decimal point.
		for(int i=startPoint; i>=offset; i--)
		{
			byte b = buffer[i];
			//System.out.println((char)b);
			if(b == '.')
			{
				pointPosition = i;
				break;
			}
		}

		// Determine if the decimal point has more numbers to the right of it then scale.
		int dif = (startPoint-pointPosition)-scale;

		//If it does change the startpoint to drop off the unwanted precision. And check the number to the right of the new startpoint to see if we need to round up.
		if(dif > 0)
		{
			startPoint = startPoint-dif;
			if(buffer[startPoint+1] > 52)
				roundUp = true;
		}

		int power = 1;

		for(int i=startPoint; i>=offset; i--)
		{
			int b = buffer[i];
			if(b > 47 && b < 58)
			{
				num = num+((b-48)*power);
			}
			else if(b == 45)
			{
				negative = true;
			}
			else if(b == 46)
			{
				//Skip the decimal
				continue;
			}
			else
			{
				byte[] bad = {(byte)b};
				throw new NumberFormatException("NumberFormatException:"+new String(bad));
			}

			power = power*10;
		}


		if(dif > 0)
		{
			//If precision was cut then check if rounding is needed.
			if(roundUp && round)
			{
				num++;
			}
		}
		else if(dif < 0)
		{
			//If precision was not high enough pad zeros on the right to reach proper precision.
			int multiplyBy = 1;
			for(int i=dif; i<0; i++)
			{
				multiplyBy = multiplyBy*10;
			}

			num = num * multiplyBy;
		}

		if(negative)
		{
			num = -num;
		}

		return num;
	}

	public long getLong() throws FunctionException
	{
		ByteArray bytes = bytesFunc.getBytes();
		int offset = bytes.getOffset();
		int numBytes = bytes.getLength();
		byte[] buffer = bytes.getBuffer();
		boolean negative = false;
		boolean roundUp = false;
		int pointPosition = 0;
		long num = 0L;

		int startPoint = pointPosition = offset+numBytes-1;
		//Scan and find the decimal point.
		for(int i=startPoint; i>=offset; i--)
		{
			byte b = buffer[i];
			//System.out.println(b);
			if(b == '.')
			{
				pointPosition = i;
				break;
			}
		}

		// Determine if the decimal point has more numbers to the right of it then precision.
		int dif = (startPoint-pointPosition)-scale;

		//If it does change the startpoint to drop off the unwanted precision. And check the number to the right of the new startpoint to see if we need to round up.
		if(dif > 0)
		{
			startPoint = startPoint-dif;
			if(buffer[startPoint+1] > 52)
			{
				roundUp = true;
			}
		}

		long power = 1L;

		for(int i=startPoint; i>=offset; i--)
		{
			int b = buffer[i];
			if(b > 47 && b < 58)
			{
				num = num+((b-48)*power);
			}
			else if(b == 45)
			{
				negative = true;
			}
			else if(b == 46)
			{
				//Skip the decimal
				continue;
			}
			else
			{
				byte[] bad = {(byte)b};
				throw new NumberFormatException("NumberFormatException:"+new String(bad));
			}

			power = power*10L;
		}

		if(dif > 0)
		{
			//If precision was cut then check if rounding is needed.
			if(roundUp && round)
			{
				num++;
			}
		}
		else if(dif < 0)
		{
			//If precision was not high enough pad zeros on the right to reach proper precision.
			int multiplyBy = 1;
			for(int i=dif; i<0; i++)
			{
				multiplyBy = multiplyBy*10;
			}

			num = num * multiplyBy;
		}

		if(negative)
		{
			num = -num;
		}

		return num;
	}
}