/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.merge;

import amberstone.system.*;
import amberstone.file.FilePath;
import amberstone.merge.*;

import java.io.*;
import java.util.*;
import java.math.*;

public class Merger
{
	private Config config;

	public Merger(Config config)
	{
		this.config = config;
	}

	public void merge() throws Exception
	{
		String s = Config.AMBERSTONE_HOME+File.separatorChar+"aggregates"+File.separatorChar+"cubes";
		String p = Config.AMBERSTONE_HOME+File.separatorChar+"aggregates"+File.separatorChar+"merged";

		File in = new File(s);
		File out = new File(p);
		walk(in, out);
	}

	private void walk(File in, File out) throws Exception
	{
		File[] summaries = getFiles(in);
		File[] dirs = getDirs(in);

		Map groups = new HashMap();

		for(int i=0; i<summaries.length; i++)
		{
			String groupKey = summaries[i].getName().split("\\.")[0];
			if(groups.containsKey(groupKey))
			{
				List group = (List)groups.get(groupKey);
				group.add(summaries[i]);
			}
			else
			{
				List group = new ArrayList();
				group.add(summaries[i]);
				groups.put(groupKey, group);
			}
		}

		Iterator it = groups.keySet().iterator();
		while(it.hasNext())
		{
			String groupKey = (String)it.next();
			List group = (List)groups.get(groupKey);
			File f = new File(out.getPath()+File.separatorChar+groupKey+".tsv");
			boolean mod = false;
			for(int i=0; i<group.size(); i++)
			{
				File summary = (File)group.get(i);
				if(f.lastModified() < summary.lastModified())
				{
					mod = true;
					break;
				}
			}

			if(mod)
			{
				merge(group, groupKey, out);
			}
		}

		for(int i=0; i<dirs.length; i++)
		{
			walk(dirs[i], new File(out, dirs[i].getName()));
		}
	}

	private void merge(List files, String groupKey, File outDir) throws Exception
	{
		String[] rows = new String[files.size()];
		BufferedReader[] ins = new BufferedReader[files.size()];
		for(int i=0; i<files.size();  i++)
		{
			File f = (File)files.get(i);
			ins[i] = new BufferedReader(new FileReader(f));
		}

		outDir.mkdirs();
		File ofile = new File(outDir, groupKey+".temp");
		PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(ofile)));

		LinkedList stack = new LinkedList();

		int t = 0;
		while(true)
		{
			++t;
			for(int i=0; i<rows.length; i++)
			{
				if(rows[i] == null)
				{
					rows[i] = ins[i].readLine();
				}
			}

			String current = null;
			String lowest = null;
			int lowestIndex = 0;

			for(int i=0; i<rows.length; i++)
			{
				current = rows[i];
				if(lowest == null)
				{
					lowest = current;
					lowestIndex = i;
				}
				else if(current != null)
				{
					//System.out.println("Compare:"+s1+" : "+s2);
					int comp = comp(lowest, current);
					if(comp < 0)
					{
						//System.out.println("Lower:"+s2);
						lowest = current;
						lowestIndex = i;
					}
				}
			}

			if(lowest == null)
			{
				break;
			}

			out.println(lowest);
			rows[lowestIndex] = null;
		}

		for(int i=0; i<ins.length; i++)
		{
			ins[i].close();
		}

		while(stack.size() > 0)
		{
			String sout = (String)stack.removeLast();
			out.println(sout);
		}

		out.close();

		BufferedReader fin = new BufferedReader(new FileReader(ofile));
		File tfile = new File(outDir, groupKey+".tsv");

		PrintWriter fout = new PrintWriter(new FileWriter(tfile));

		String current = null;
		String last = null;
		while((current = fin.readLine()) != null)
		{
			if(last != null)
			{
				if(comp(last, current) == 0)
				{
					current = merge(last, current);
				}
				else
				{
					fout.println(last.replace("::", "	"));
				}
			}

			last = current;
		}

		fout.println(last.replace("::", "	"));
		fin.close();
		fout.close();
		ofile.delete();
	}

	private String merge(String s1, String s2)
	{
		String parts1[] = s1.split("	");
		String parts2[] = s2.split("	");
		for(int i=1; i<parts1.length; i++)
		{
			long l1 = Long.parseLong(parts1[i]);
			long l2 = Long.parseLong(parts2[i]);
			parts1[i]  = Long.toString(l1+l2);
		}

		return join(parts1);
	}

	private String join(String parts[])
	{
		StringBuffer buf = new StringBuffer();
		for(int i=0; i<parts.length; i++)
		{
			if(i > 0)
			{
				buf.append("	");
			}

			buf.append(parts[i]);
		}

		return buf.toString();
	}

	private int comp(String s1, String s2)
	{
		String parts1[] = s1.split("	");
		String parts2[] = s2.split("	");
		return parts1[0].compareTo(parts2[0]);
	}

	private File[] getDirs(File root)
	{
		ArrayList<File> dirs = new ArrayList();
		File[] files = root.listFiles();
		for(int i=0;i<files.length; i++)
		{
			if(files[i].isDirectory())
			{
				dirs.add(files[i]);
			}
		}

		return dirs.toArray(new File[dirs.size()]);
	}

	private File[] getFiles(File root)
	{
		ArrayList<File> dirs = new ArrayList();
		File[] files = root.listFiles();
		for(int i=0;i<files.length; i++)
		{
			if(files[i].getName().endsWith(".tsv"))
			{
				dirs.add(files[i]);
			}
		}

		return dirs.toArray(new File[dirs.size()]);
	}
}