/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.resources;

import amberstone.system.SystemException;

import jdbm.RecordManager;
import jdbm.RecordManagerFactory;
import jdbm.helper.FastIterator;
import jdbm.htree.HTree;

import amberstone.lang.ResourceFunction;
import amberstone.util.LRUCache;
import amberstone.data.ByteArray;

import java.io.File;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.Map;
import java.util.HashMap;
import java.util.Properties;
import java.sql.*;


public class DiskResource implements Resource,
									 ResourceFunction
{

	private HTree index;
	private HTree reverseIndex;
	private RecordManager recman;
	private String dir;
	private String name;

	public DiskResource(String dir, String name)
	{
		this.dir = dir;
		this.name = name;
	}

	public Resource getResource()
	{
		return this;
	}

	public void load() throws SystemException
	{
		File resourcesFile = new File(dir+File.separatorChar+name+".txt");
		if(resourcesFile.exists())
		{
			loadFile(resourcesFile);
			return;
		}

		resourcesFile = new File(dir+File.separatorChar+name+".sql");
		if(resourcesFile.exists())
		{
			loadSQL(resourcesFile);
			return;
		}

		throw new SystemException("Resource file not found: "+resourcesFile.getPath());
	}

	public void loadFile(File resourcesFile)
		throws SystemException
	{
		try
		{
			if(!exists() || lastModified() < resourcesFile.lastModified())
			{

				if(exists())
				{
					clean();
				}

				open();

				BufferedReader in = null;
				try
				{
					in = new BufferedReader(new FileReader(resourcesFile));
					int count = 0;
					for(String line = in.readLine(); line != null; line = in.readLine())
					{
						line = line.trim();
						if(line.indexOf(":") > -1 && line.indexOf("#") != 0)
						{
							int delimIndex = line.indexOf(":");

							String name  = line.substring(0, delimIndex).trim();
							String value = line.substring(delimIndex+1).trim();

							put(new ByteArray(name), new ByteArray(value));
							++count;
							if(count % 5000 == 0)
							{
								commit();
							}
						}
					}
				}
				finally
				{
					commit();
				}
			}
			else
			{
				open();
			}
		}
		catch(Exception e)
		{
			throw new SystemException("Error loading file resource:"+name, e);
		}
	}

	private void loadProps(File f, Properties p) throws Exception
	{
		BufferedReader in = null;
		try
		{
			in = new BufferedReader(new FileReader(f));
			for(String line = in.readLine(); line != null; line = in.readLine())
			{
				line = line.trim();
				if(line.indexOf(":") > -1 && line.indexOf("#") != 0)
				{
					int delimIndex = line.indexOf(":");
					String name  = line.substring(0, delimIndex).trim();
					String value = line.substring(delimIndex+1).trim();
					p.put(name, value);
				}
			}
		}
		finally
		{
			in.close();
		}
	}

	public void loadSQL(File resourcesFile) throws SystemException
	{
		Properties props = new Properties();
		try
		{
			loadProps(resourcesFile, props);
		}
		catch(Exception e)
		{
			throw new SystemException("Error load SQL props:"+name, e);
		}

		String driver   = props.getProperty("driver");
		String url      = props.getProperty("url");
		String query    = props.getProperty("query");
		String username = props.getProperty("username");
		String password = props.getProperty("password");
		int reload      = Integer.parseInt(props.getProperty("reload", "86400"));

		if(driver == null)
			throw new SystemException("Error loading SQL resource: driver undefined");

		if(url == null)
			throw new SystemException("Error loading SQL resource: url undefined");

		if(query == null)
			throw new SystemException("Error loading SQL resource: query undefined");

		if(username == null)
			throw new SystemException("Error loading SQL resource: username undefined");

		if(password == null)
			throw new SystemException("Error loading SQL resource: password undefined");

		if(!exists() || (System.currentTimeMillis()/1000) - lastModified() > reload)
		{
			Connection con = null;
			PreparedStatement stmt = null;
			ResultSet rs = null;
			try
			{
				if(exists())
				{
					clean();
				}

				open();

				Class.forName(driver);
				con = DriverManager.getConnection(url, username, password);
				stmt = con.prepareStatement(query);
				rs = stmt.executeQuery();
				int count = 0;
				while(rs.next())
				{
					String key   = rs.getString("key");
					String value = rs.getString("value");
					put(new ByteArray(key), new ByteArray(value));
					++count;
					if(count % 5000 == 0)
					{
						commit();
					}
				}

				commit();
			}
			catch(Exception e)
			{
				throw new SystemException("Error loading SQL resource", e);
			}
			finally
			{
				try
				{
					rs.close();
				}
				catch(Exception e1)
				{
					e1.printStackTrace();
				}

				try
				{
					stmt.close();
				}
				catch(Exception e1)
				{
				}

				try
				{
					con.close();
				}
				catch(Exception e1)
				{
					e1.printStackTrace();
				}
			}
		}
		else
		{
			try
			{
				open();
			}
			catch(Exception e)
			{
				throw new SystemException("Error opening resource:"+name, e);
			}
		}
	}

	public void open() throws IOException
	{
		Properties props = new Properties();
		props.put("jdbm.disableTransactions", "TRUE");

		recman = RecordManagerFactory.createRecordManager(this.dir+File.separatorChar+"db"+File.separatorChar+this.name, props);

		long recid = recman.getNamedObject("index");
		if (recid != 0)
		{
			this.index = HTree.load(recman, recid);
		}
		else
		{
			this.index = HTree.createInstance(recman);
			recman.setNamedObject("index", index.getRecid());
		}

		recid = recman.getNamedObject("reverseIndex");
		if (recid != 0)
		{
			this.reverseIndex = HTree.load(recman, recid);
		}
		else
		{
			this.reverseIndex = HTree.createInstance(recman);
			recman.setNamedObject("reverseIndex", reverseIndex.getRecid());
		}
	}

	public void clean() throws IOException
	{
		File db = new File(this.dir+File.separatorChar+"db"+File.separatorChar+this.name+".db");
		File log = new File(this.dir+File.separatorChar+"db"+File.separatorChar+this.name+".lg");
		db.delete();
		log.delete();
	}

	public ResourceIterator keys() throws IOException
	{
		return new DiskResourceIterator(index.keys());
	}

	public ResourceIterator values() throws IOException
	{
		return new DiskResourceIterator(reverseIndex.keys());
	}

	public String getName()
	{
		return this.name;
	}

	public String getDir()
	{
		return this.dir;
	}

	public long lastModified()
	{
		File db = new File(this.dir+File.separatorChar+"db"+File.separatorChar+this.name+".db");
		return db.lastModified();
	}

	public void close() throws IOException
	{
		this.recman.commit();
		this.recman.close();
	}

	public boolean exists()
	{
		File db = new File(this.dir+File.separatorChar+"db"+File.separatorChar+this.name+".db");
		return db.exists();
	}

	public void commit() throws IOException
	{
		recman.commit();
	}

	public String get(String key)
		throws IOException
	{
		ByteArray value = getValue(new ByteArray(key));
		if(value != null)
		{
			return value.toString();
		}

		return null;
	}

	public ByteArray getValue(ByteArray key)
		throws IOException
	{
		return (ByteArray)index.get(key);
	}

	public ByteArray getKey(ByteArray value)
		throws IOException
	{
		return (ByteArray)reverseIndex.get(value);
	}

	public void put(ByteArray key, ByteArray value)
		throws IOException
	{
		index.put(key, value);
		reverseIndex.put(value, key);
	}
}