/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.summary;

import amberstone.file.FileStat;
import amberstone.data.Key;
import amberstone.data.ByteArray;
import amberstone.lang.*;
import amberstone.lang.FunctionException;
import amberstone.merge.Merge;
import amberstone.merge.AddMerge;
import amberstone.util.Primes;
import amberstone.time.TimeSeries;
import amberstone.system.*;
import amberstone.report.ActiveKey;

import java.io.File;
import java.io.PrintWriter;
import java.io.FileWriter;
import java.io.IOException;

import java.util.*;

public class Cube
	implements FileNode
{

	private String name;
	private KeyFunction keyFunc;
	private Point[] points;
	private int view;
	private int maxSize;
	private int buckets;
	private int resize;
	private int count;
	private long median;
	private int acount;
	private SummaryFunction operator;
	private Comparator pc = new PointComp();

	private int increment;
	private BooleanFunction filter;
	private BytesFunction bytesFunc;
	private long grain;

	public Cube(String name,
			    int maxSize,
			    SummaryFunction operator,
				KeyFunction keyFunc,
				BytesFunction bytesFunc,
				BooleanFunction filter)
	{
		this.name = name;
		this.keyFunc = keyFunc;
		this.maxSize = maxSize;
		this.operator = operator;
		this.bytesFunc = bytesFunc;
		this.filter = filter;
		points = new Point[113];
		resize = (int)(points.length*.65);
	}

	public void output()
		throws Exception
	{
		PrintWriter out = null;
		try
		{
			String fullPath = Summarizer.outDir.toString();
			String grainDim = null;
			if(grain > 2147483647)
			{
				int dayGrain = (int)(grain >>> 32);
				int hourGrain = (int)grain;
				grainDim = formatDayGrain(dayGrain)+"::"+hourGrain;
				fullPath = fullPath+File.separatorChar+dayGrain+File.separatorChar+hourGrain;
			}
			else
			{
				int dayGrain = (int)grain;
				grainDim = formatDayGrain(dayGrain);
				fullPath = fullPath+File.separatorChar+dayGrain;
			}

			File odir = new File(fullPath);
			odir.mkdirs();
			File outFile = new File(odir, name+"."+SummarizeCommand.index+"."+System.getenv("AMBERSTONE_NODE")+".tsv");
			out = new PrintWriter(new FileWriter(outFile));

			LinkedList ll = getPoints();
			Collections.sort(ll, pc);

			int values = operator.getValues();

			while(ll.size() > 0)
			{
				Point point = null;
				point = (Point)ll.removeFirst();
				String s = point.key.toString();
				out.print(grainDim+"::"+s);
				out.print("	");
				out.print(point.count);
				if(values > 1)
				{
					out.print("	");
					out.print(point.count2);
				}

				if(values > 2)
				{
					out.print("	");
					out.print(point.count3);
				}
				out.println("");
			}
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		finally
		{
			out.close();
		}
	}

	public String formatDayGrain(int dayGrain)
	{
		String s = Integer.toString(dayGrain);

		if(s.length() == 8)
		{
			s = formatYYYYMMDD(s);
		}
		else if(s.length() == 7)
		{
			s = formatYYYYWWD(s);
		}

		return s;
	}

	public String join(String[] parts)
	{
		String r = "";
		for(int i=0; i<parts.length; i++)
		{
			if(i > 0)
			{
				r = r + "::";
			}
			r = r + parts[i];

		}

		return r;
	}

	public String formatYYYYWWD(String YYYYWWD)
	{
		StringBuilder buf = new StringBuilder();
		buf.append(YYYYWWD.substring(0,4)).append("::").append(YYYYWWD.substring(4,6)).append("::").append(YYYYWWD.substring(6));
		return buf.toString();
	}

	public String formatYYYYMMDD(String YYYYMMDD)
	{
		StringBuilder buf = new StringBuilder();
		buf.append(YYYYMMDD.substring(0,4)).append("::").append(quarter(YYYYMMDD.substring(4,6))).append("::").append(YYYYMMDD.substring(4,6)).append("::").append(YYYYMMDD.substring(6));
		return buf.toString();
	}

	public String quarter(String MM)
	{
		int m = Integer.parseInt(MM);
		switch(m)
		{
			case 1:
				return "1";
			case 2:
				return "1";
			case 3:
				return "1";
			case 4:
				return "2";
			case 5:
				return "2";
			case 6:
				return "2";
			case 7:
				return "3";
			case 8:
				return "3";
			case 9:
				return "3";
			case 10:
				return "4";
			case 11:
				return "4";
			case 12:
				return "4";
		}

		return "0";
	}

	public String getName()
	{
		return this.name;
	}

	public void finish(long grain)
	{
		this.grain = grain;
	}

	private void push(Point point) throws FunctionException
	{
		operator.operate(point);
		count++;

		if(count > resize)
		{
			resize();
		}
	}

	private void resize()
	{
		LinkedList<Point> ll = getPoints();
		int s = (int)(points.length*1.75);
		s = Primes.selectPrimeCandidate(s);
		resize = (int)(s*.65);
		points = new Point[s];


		while(ll.size() > 0)
		{
			addToHash(ll.removeLast());
		}
	}

	private void addToHash(Point point)
	{
		point.link = null;
		Key key = point.key;

		int index = key.hashCode() % this.points.length;
		Point p = points[index];

		if(p == null)
		{
			points[index] = point;
		}
		else
		{
			while(p.link != null)
			{
				p = p.link;
			}

			p.link = point;
		}
	}

	private LinkedList<Point> getPoints()
	{
		LinkedList pointList = new LinkedList();
		for(int i=0; i<points.length; i++)
		{
			for(Point point = points[i]; point != null; point = point.link)
			{
				pointList.addLast(point);
			}
		}

		return pointList;
	}

	class PointComp implements Comparator
	{
		public int compare(Object o1, Object o2)
		{
			Point p1 = (Point)o1;
			Point p2 = (Point)o2;
			return p2.key.toString().compareTo(p1.key.toString());
		}
	}

	public void update() throws FunctionException
	{
		if(!filter.getBoolean())
		{
			return;
		}

		Key key = keyFunc.getKey();

		int index = key.hashCode() % this.points.length;

		Point p = null;
		Point lastPoint = null;

		for(p = points[index]; p != null; lastPoint = p, p = p.link)
		{
			if(key.equalsKey(p.key))
			{
				operator.operate(p);
				return;
			}
		}

		if(count == maxSize)
		{
			return;
		}

		p = new Point();
		operator.operate(p);
		p.key = key.cloneKey();

		if(lastPoint != null)
		{
			lastPoint.link = p;
		}
		else
		{
			points[index] = p;
		}

		++count;
		if(count > resize)
		{
			resize();
		}
	}
}