/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.summary;

import java.io.FileFilter;
import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.TreeSet;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Calendar;
import java.util.GregorianCalendar;

import amberstone.system.*;
import amberstone.time.*;
import amberstone.data.DataRecord;
import amberstone.data.ByteArray;
import amberstone.data.Field;
import amberstone.file.FileStat;
import amberstone.file.FilePath;
import amberstone.dictionary.Dictionary;
import amberstone.read.Blocks;


public class Summarizer
{
	private Config config;
	private RangeFilter rangeFilter;
	private int maxFileSize;
	private SummaryThread[] summaryThreads;
	private int logLevel;
	private Logger logger;
	private String summariesHome;
	private String dataWarehouseHome;
	private int startYYYYMMDD;
	private int endYYYYMMDD;
	private int startBlock;
	private String runFile;
	public static String outDir;
	public Calendar cal = new GregorianCalendar();

	public Summarizer(Config config,
					  int startYYYYMMDD,
					  int endYYYYMMDD,
					  int startBlock)
	{
		this.startYYYYMMDD = startYYYYMMDD;
		this.endYYYYMMDD = endYYYYMMDD;
		this.startBlock = startBlock;
		this.config = config;
		outDir = config.AMBERSTONE_HOME+File.separatorChar+"aggregates"+File.separatorChar+"cubes";
	}

	public void summarize() throws Exception
	{
		try
		{

			this.logLevel       = Integer.parseInt(config.getProperty(Config.LOG_LEVEL));
			int logMode         = Integer.parseInt(config.getProperty(Config.LOG_MODE));

			IndexConfig indexConfig             = config.getIndexConfig();
			SummarizersConfig summarizersConfig = config.getSummarizersConfig();

			this.dataWarehouseHome      = indexConfig.getIndexDir();

			DataRecord dataRecord = indexConfig.getDataRecord();

			String logsHome       = config.getLogsHome();
			this.logger = new Logger(logsHome, "aggregator", logMode);
			this.logger.timestamp();
			this.logger.log("Aggregator startup");
			this.summaryThreads = summarizersConfig.getSummaryThreads();
			this.summaryThreads = selectThreads(this.summaryThreads);
			List<Field> activeFields = selectActiveFields(this.summaryThreads);
			this.maxFileSize = Integer.parseInt(config.getProperty(Config.INDEX_BLOCK_SIZE));
			logger.ln();
			logger.log("Aggregator settings");
			logger.log("-"+Config.INDEX_BLOCK_SIZE+": "+this.maxFileSize);
			logger.log("-Partition: "+ dataWarehouseHome);

			Blocks blocks = new Blocks();
			int blockCount = blocks.count(dataWarehouseHome, startYYYYMMDD, endYYYYMMDD, startBlock);
			if(blockCount == 0)
			{
				System.out.println("0 records found.");
				return;
			}
			else
			{
				System.out.println("Scanning "+blockCount+" blocks");
			}

			ReadAheadThread readAheadThread = new ReadAheadThread(dataWarehouseHome,
															      startYYYYMMDD,
															      endYYYYMMDD,
															      startBlock,
															      maxFileSize,
															      dataRecord,
															      activeFields);
			WriterThread writerThread = new WriterThread();
			writerThread.start();

			readAheadThread.setLogInfo(logsHome, logMode, logLevel);
			readAheadThread.start();

			logger.ln();
			for(int i=0; i<summaryThreads.length; i++)
			{
				this.logger.log("Initailizing thread: "+i);
				summaryThreads[i].setThreadNumber(i);
				summaryThreads[i].start();
			}

			if(summarizeFiles(readAheadThread, blockCount))
			{
				advanceDay();
				for(int i=0; i<summaryThreads.length; i++)
				{
					summaryThreads[i].finish();
				}

			}
			WriterThread.put(new String(""));
		}
		catch(SystemException se)
		{
			se.printStackTrace();
			WriterThread.put(new String(""));
			logger.log("Exception occurred during aggregation:");
			se.log(logLevel, logger);
			throw se;
		}
		finally
		{
			this.logger.ln();
			this.logger.timestamp();
			this.logger.log("Aggregator shutdown");
			logger.close();
		}
	}

	private void advanceDay()
	{
		int yyyy = ActiveYear.ACTIVE_YEAR.intValue();
		int mm = ActiveMonth.ACTIVE_MONTH.intValue();
		int dd = ActiveDay.ACTIVE_DAY.intValue();

		cal.set(yyyy, mm-1, dd, 0, 0, 0);
		cal.add(cal.DAY_OF_MONTH, 1);

		yyyy = cal.get(cal.YEAR);
		mm = cal.get(cal.MONTH)+1;
		dd = cal.get(cal.DAY_OF_MONTH);
		ActiveYearMonthDay.ACTIVE_KEY.setInt(((yyyy*10000)+(mm*100)+dd));
	}

	private SummaryThread[] selectThreads(SummaryThread[] threads)
	{
		List<SummaryThread> list = new ArrayList();
		for(int i=0; i<threads.length; i++)
		{
			if(threads[i].getFileTemplates().length > 0)
			{
				list.add(threads[i]);
			}
		}

		return (SummaryThread[])list.toArray(new SummaryThread[list.size()]);
	}

	private List<Field> selectActiveFields(SummaryThread[] threads)
	{
		List<Field> list = new ArrayList();
		for(int i=0; i<threads.length; i++)
		{
			DataRecord record = threads[i].record;
			Field[] fields = record.getActiveFields();
			for(int f=0; f<fields.length; f++)
			{
				if(!list.contains(fields[f]))
				{
					list.add(fields[f]);
				}
			}
		}

		return list;
	}

	private String pad(int i)
	{
		String s = Integer.toString(i);
		s= s+"%";
		while(s.length() < 5)
		{
			s = s+" ";
		}

		return s;
	}

	private int currentDay;

	private boolean summarizeFiles(ReadAheadThread readAheadThread, int numBlocks)
		throws Exception
	{
		long begin = 0L;
		long end = 0L;
		System.out.print("Completion status: 0%   ");

		int blockCount = 0;
		for(DataRecord currentRecord = readAheadThread.getReadyRecord(); currentRecord != null; currentRecord = readAheadThread.getReadyRecord())
		{
			++blockCount;

			logger.ln();
			logger.log("Summarizing block: "+currentRecord.getDir().getPath());
			begin = System.nanoTime();

			parseDay(currentRecord.getDir());

			for(int i=0; i<summaryThreads.length; i++)
			{
				summaryThreads[i].putDataRecord(currentRecord);
			}

			for(int i=0; i<summaryThreads.length; i++)
			{
				DataRecord finishedRecord = summaryThreads[i].getDataRecord();
				if(finishedRecord.e != null)
				{
					finishedRecord.e.printStackTrace();
					System.out.println("");
					System.out.println("");
					System.out.println("Exception in summary thread "+i);
					System.out.println("");
					System.out.println(finishedRecord.e.getMessage());
					System.out.println("Summarization aborted. See thread log for details.");

					logger.log("Exception in summary thread "+i);

					readAheadThread.putFinishedRecord(finishedRecord);
					for(int j=0; j<summaryThreads.length; j++)
					{
						System.out.println("Signaling shutdown to thread: "+j);
						logger.log("Signaling shutdown to thread: "+j);
						summaryThreads[j].putDataRecord(finishedRecord);
					}

					return false;
				}
			}

			readAheadThread.putFinishedRecord(currentRecord);
			end = System.nanoTime();
			int percentComplete =(int)(((double)((double)blockCount/(double)numBlocks)) * 100D);
			System.out.print("\b\b\b\b\b");
			System.out.print(pad(percentComplete));
			logger.log("Time: "+Double.toString((double)(end-begin)/(double)1000000));
		}


		logger.ln();
		System.out.println("");

		for(int i=0; i<summaryThreads.length; i++)
		{
			logger.log("Signaling shutdown to thread: "+i);
			summaryThreads[i].putDataRecord(new DataRecord());
		}

		return true;
	}

	private DaySeries daySeries = new DaySeries();

	private void parseDay(File block) throws Exception
	{
		File day = block.getParentFile();
		int dd = Integer.parseInt(day.getName());
		File month = day.getParentFile();
		int mm = Integer.parseInt(month.getName());
		File year = month.getParentFile();
		int yyyy = Integer.parseInt(year.getName());

		int YYYYMMDD = (yyyy*10000)+(mm*100)+dd;

		if(YYYYMMDD != currentDay)
		{
			currentDay = YYYYMMDD;
			ActiveMonth.ACTIVE_MONTH.setInt(mm);
			ActiveDay.ACTIVE_DAY.setInt(dd);
			ActiveYear.ACTIVE_YEAR.setInt(yyyy);
			ActiveYearMonthDay.ACTIVE_KEY.setInt(YYYYMMDD);

			int q = daySeries.getQuarter(mm);

			int ww = daySeries.getWeek(YYYYMMDD);

			int weekYear = yyyy;
			if(mm == 1 && ww > 6)
			{
				weekYear = yyyy-1;
			}

			int dayOfWeek = daySeries.getDayOfWeek(YYYYMMDD);

			int YYYYQQMMDD = (yyyy*100000)+(q*10000)+(mm*100)+dd;
			int YYYYWWD     = (weekYear*1000)+(ww*10)+dayOfWeek;

			ActiveYearQuarterMonthDay.ACTIVE_KEY.setInt(YYYYQQMMDD);
			ActiveYearWeekDay.ACTIVE_KEY.setInt(YYYYWWD);
		}
	}
}
