/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.system;

import amberstone.data.ByteArray;
import amberstone.lang.Function;
import amberstone.load.Loader;
import amberstone.read.Reader;
import amberstone.dictionary.Dictionary;
import amberstone.dictionary.DiskDictionary;
import amberstone.resources.Resource;
import amberstone.resources.DiskResource;
import amberstone.file.FilePath;
import amberstone.summary.Summarizer;

import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.Iterator;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.File;
import java.io.IOException;
import java.util.regex.Pattern;

public class CommandRunner
{
	public static void main(String[] args) throws Exception
	{
		System.out.println("");
		CommandRunner cr = new CommandRunner(args);
	}

	public CommandRunner(String[] args)
	{

		Command com = getCommand(args);

		if(!com.valid())
		{
			System.out.println(com.usage());
			return;
		}

		Config config = null;
		try
		{
			config = new Config();
		}
		catch(Throwable e)
		{
			System.out.println("Initialization error:");
			System.out.println(e.getMessage());
			Trace trace = new Trace();
			trace.trace(e, "system", config.getLogsHome());
			return;
		}

		try
		{
			config.parse(com);
		}
		catch(Throwable e)
		{
			Trace trace = new Trace();
			trace.trace(e, "system", config.getLogsHome());
			return;
		}


		com.run(config);
	}

	private Command getCommand(String[] args)
	{
		String c = args[0];
		if(c.equals("-summarize"))
		{
			return new SummarizeCommand(args);
		}
		else if(c.equals("-load"))
		{
			return new LoadCommand(args);
		}
		else if(c.equals("-read"))
		{
			return new ReadCommand(args);
		}
		else if(c.equals("-delete"))
		{
			return new DeleteCommand(args);
		}
		else if(c.equals("-dump"))
		{
			return new DumpCommand(args);
		}
		else if(c.equals("-export"))
		{
			return new ExportCommand(args);
		}
		else if(c.equals("-trunc"))
		{
			return new TruncCommand(args);
		}
		else if(c.equals("-part"))
		{
			return new PartCommand(args);
		}
		else if(c.equals("-merge"))
		{
			return new MergeCommand(args);
		}
		else if(c.equals("-rollup"))
		{
			return new RollupCommand(args);
		}
		else if(c.equals("-archive"))
		{
			return new ArchiveCommand(args);
		}
		else if(c.equals("-stat"))
		{
			return new StatCommand(args);
		}
		else if(c.equals("-drop"))
		{
			return new DropCommand(args);
		}

		return null;
	}

	private static void openResources(Config config) throws Exception
	{
		Map<String, Resource> resources = config.getResources();

		Iterator<Resource> it = resources.values().iterator();

		List<DiskResource> opened = new ArrayList();

		while(it.hasNext())
		{
			DiskResource dr = (DiskResource)it.next();

			try
			{
				dr.load();
				opened.add(dr);
			}
			catch(Exception e)
			{
				for(int i=0; i<opened.size(); i++)
				{
					try
					{
						opened.get(i).close();
					}
					catch(Exception e1)
					{

					}
				}
				throw e;
			}
		}
	}

	private static void openDictionaries(Config config) throws Exception
	{
		Map<String, Dictionary> dicts = config.getDictionaries();

		Iterator it = dicts.values().iterator();

		List<DiskDictionary> opened = new ArrayList();

		while(it.hasNext())
		{
			DiskDictionary dd = (DiskDictionary)it.next();
			try
			{
				dd.open();
				opened.add(dd);
			}
			catch(Exception e)
			{
				for(int i=0; i<opened.size(); i++)
				{
					try
					{
						opened.get(i).close();
					}
					catch(Exception e1)
					{

					}
				}
				throw e;
			}
		}
	}

	private static void printUsage()
	{
		System.out.println("Usage:");
		System.out.println("stone [-options]");
		System.out.println("options include:");
		System.out.println("-delete <fields|summaries|reports> [regex]   Drops matching objects");
		System.out.println("-clean  Drops all data");
		System.out.println("-list <summaries|reports> <regex>   Lists matching objects");
		System.out.println("-dump <dictionary|resource> <name> [key]   Drops matching objects");
		System.out.println("-load        Runs the loader");
		System.out.println("-archive     Archives load files");
		System.out.println("-read        Runs the reader");
		System.out.println("-extract     Runs the reader in extract mode");
		System.out.println("-summarize   Runs the summarizer");
		System.out.println("-accumulate  Runs the accumulator");
		System.out.println("-consolidate  Runs the consolidator");
		System.out.println("-alert       Runs the alerter");
	}
}