/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.system;

import amberstone.data.*;
import amberstone.data.fields.*;

import amberstone.lang.*;
import amberstone.read.*;
import amberstone.load.*;
import amberstone.dictionary.*;
import amberstone.summary.*;
import amberstone.transaction.*;
import amberstone.time.*;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.ErrorHandler;
import org.xml.sax.Locator;
import org.xml.sax.helpers.DefaultHandler;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import java.util.LinkedList;
import java.util.ArrayList;
import java.util.Properties;
import java.io.IOException;
import java.io.File;

public class IndexConfig extends DefaultHandler implements ErrorHandler
{
	private Config parent;
	private int fieldIndex;
	private Locator locator;
	private amberstone.lang.Compiler compiler;
	private List fields = new ArrayList();
	private DataRecord dataRecord;
	private String loadDir;
	private String indexDir;
	private String summariesDir;
	public static String index;
	private Command command;
	private Properties system;
	private String AMBERSTONE_HOME;
	private amberstone.read.Selector select;
	private Set activeFields = new TreeSet();
	private TreeSet activeDelimitedFields = new TreeSet();

	private DelimitedRecord delimitedRecord;
	private Meter loadMeter;
	private BooleanFunction loadFilter;
	private boolean showLocation = true;

	private Map<String, Dictionary> dictionaries = new HashMap();
	private List<PropTemplate> tprops = new ArrayList();
	private FileTemplate currentFileTemplate;
	private Meter readMeter;
	private String dictionariesDir;
	private int commitEvery;
	private boolean transactional;


	public IndexConfig(Config parent,
					   Command command,
	                   amberstone.lang.Compiler compiler,
	                   Properties system)
		throws Exception
	{
		this.AMBERSTONE_HOME = AmberstoneConfig.AMBERSTONE_HOME;
		this.parent = parent;
		this.command = command;
		this.compiler = compiler;
		this.system = system;
		this.commitEvery = Integer.parseInt(system.getProperty(Config.DICTIONARIES_COMMIT_EVERY));
		this.transactional = Boolean.parseBoolean(system.getProperty(Config.DICTIONARIES_TRANSACTIONAL));
		DiskDictionary.commitEvery = commitEvery;
		DiskDictionary.transactions = transactional;
		DictionaryFactory.setDictionaries(this.dictionaries);
		String config = null;
		if(command instanceof LoadCommand)
		{
			LoadCommand loadCommand = (LoadCommand)command;
			this.index = loadCommand.getIndex();

			config = AMBERSTONE_HOME + File.separatorChar + "storage" + File.separatorChar + index + File.separatorChar + "conf" + File.separatorChar + "partition.xml";
			if(loadCommand.archive())
			{
				this.loadDir = AMBERSTONE_HOME + File.separatorChar + "storage" + File.separatorChar + index + File.separatorChar + "load-archive";
			}
			else
			{
				this.loadDir = AMBERSTONE_HOME + File.separatorChar + "storage" + File.separatorChar + index + File.separatorChar + "load-queue";
			}

			this.indexDir = AMBERSTONE_HOME + File.separatorChar + "storage" + File.separatorChar + index + File.separatorChar + "table";
			DiskDictionary.dir = AMBERSTONE_HOME + File.separatorChar + "storage" + File.separatorChar + index + File.separatorChar + "dictionaries";
		}
		else if(command instanceof ReadCommand)
		{
			ReadCommand readCommand = (ReadCommand)command;
			this.index = readCommand.getIndex();
			config = AMBERSTONE_HOME + File.separatorChar + "storage" + File.separatorChar + index + File.separatorChar + "conf" + File.separatorChar + "partition.xml";
			this.indexDir = AMBERSTONE_HOME + File.separatorChar + "storage" + File.separatorChar + index + File.separatorChar + "table";
			DiskDictionary.dir = AMBERSTONE_HOME + File.separatorChar + "storage" + File.separatorChar + index + File.separatorChar + "dictionaries";
		}
		else if(command instanceof SummarizeCommand)
		{
			SummarizeCommand summarizeCommand = (SummarizeCommand)command;
			this.index = summarizeCommand.getIndex();
			config = AMBERSTONE_HOME + File.separatorChar + "storage" + File.separatorChar + index + File.separatorChar + "conf" + File.separatorChar + "partition.xml";
			this.indexDir = AMBERSTONE_HOME + File.separatorChar + "storage" + File.separatorChar + index + File.separatorChar + "table";
			DiskDictionary.dir = AMBERSTONE_HOME + File.separatorChar + "storage" + File.separatorChar + index + File.separatorChar + "dictionaries";
		}
		else if(command instanceof StatCommand)
		{
			StatCommand statCommand = (StatCommand)command;
			this.index = statCommand.getIndex();
			config = AMBERSTONE_HOME + File.separatorChar + "storage" + File.separatorChar + index + File.separatorChar + "conf" + File.separatorChar + "partition.xml";
			this.indexDir = AMBERSTONE_HOME + File.separatorChar + "storage" + File.separatorChar + index + File.separatorChar + "table";
			this.summariesDir = AMBERSTONE_HOME + File.separatorChar + "analytics" + File.separatorChar + "summaries";
			DiskDictionary.dir = AMBERSTONE_HOME + File.separatorChar + "storage" + File.separatorChar + index + File.separatorChar + "dictionaries";
		}

		parse(config);

		if(command instanceof ReadCommand)
		{
			ReadCommand readCommand = (ReadCommand)command;
			Col.setDataRecord(this.dataRecord);
			Col.setActiveFields(this.activeFields);
			Col.inContext = true;

			String sel = readCommand.getSelect();

			showLocation = false;
			try
			{
				Object o = compiler.compile(sel);
				if(!(o instanceof amberstone.read.Selector))
				{
					throw new SystemException("Invalid function type, Select expected: "+sel);
				}

				this.dataRecord.setActiveFields((Field[])this.activeFields.toArray(new Field[activeFields.size()]));
				Reader.dataRecord = this.dataRecord;
				Reader.select = (amberstone.read.Selector)o;
			}
			catch(Exception e)
			{
				System.out.println(e.getMessage());
				Throwable c = e.getCause();
				int i = 1;
				while(c != null)
				{
					System.out.println("-Cause:");
					System.out.println(c.getMessage());
					c = c.getCause();
					i++;
				}
				throw e;
			}
		}
	}

	public String getIndex()
	{
		return this.index;
	}

	public String getIndexDir()
	{
		return this.indexDir;
	}

	public String getLoadDir()
	{
		return this.loadDir;
	}

	public String getSummariesDir()
	{
		return this.summariesDir;
	}

	public Map getDictionaries()
	{
		return this.dictionaries;
	}

	public BooleanFunction getLoadFilter()
	{
		return this.loadFilter;
	}

	public Meter getLoadMeter()
	{
		return this.loadMeter;
	}

	public DelimitedRecord getDelimitedRecord()
	{
		return this.delimitedRecord;
	}

	public void parse(String path) throws Exception
	{
		XMLReader parser = (XMLReader)Class.forName("org.apache.xerces.parsers.SAXParser").newInstance();
	  	parser.setFeature("http://xml.org/sax/features/validation", true);
		parser.setContentHandler(this);
		parser.setErrorHandler(this);
		parser.parse(path);
	}

	public void setDocumentLocator(Locator locator)
	{
		this.locator = locator;
	}

	public DataRecord getDataRecord()
	{
		return this.dataRecord;
	}

	public void startElement(String uri,
						     String local,
							 String raw,
	                         Attributes attrs)
	{
		try
		{

			if(raw.equals("columns"))
			{
				int maxFields =          Integer.parseInt(system.getProperty(Config.MAX_LOAD_FIELDS));
			    int inputBufferSize =    Integer.parseInt(system.getProperty(Config.LOAD_READ_BUFFER_SIZE));

				this.delimitedRecord = new DelimitedRecord(inputBufferSize, maxFields);
				DataColumn.setDelimitedRecord(this.delimitedRecord);
				DataColumn.setActiveDelimitedFields(this.activeDelimitedFields);
				this.fields = new ArrayList();
				this.fields = new ArrayList();
			}
			else if(raw.equals("column"))
			{
				++this.fieldIndex;

				String type = attrs.getValue("type");
				String name = attrs.getValue("name");

				if(type.equals("int"))
				{
					int size = parseInt(attrs.getValue("size"), 4);
					String func = "0";
					Field ifield = getIntField(size, type, name, this.fieldIndex, func);
					String src = attrs.getValue("dictionary");
					if(src != null)
					{
						Object o = compiler.compile(src);
						if(o instanceof DictionaryFunction)
						{
							DictionaryFunction d = (DictionaryFunction)o;
							((Int)ifield).setDictionary(d.getDictionary());
						}
						else
						{
							throw new SystemException("Dictionary expected:"+src);
						}
					}

					this.fields.add(ifield);
				}
				else if(type.equals("time"))
				{
					String func = "0";
					IntFunction ifunc = (IntFunction)compiler.compile(func);
					Field ifield = new TimeInt(name, this.fieldIndex, ifunc);
					String template = attrs.getValue("template");
					((TimeInt)ifield).setTemplate(template);
					this.fields.add(ifield);
				}
				else if(type.equals("long"))
				{
					String func = "0";
					int typeValue = getTypeValue(type);
					typeValue = typeValue > -1 ? typeValue : -2;
					LongFunction lfunc = (LongFunction)compiler.compile(func);
					Field lfield = new LongInt(typeValue, name, this.fieldIndex, lfunc);
					this.fields.add(lfield);
				}
				else if(type.equals("bytes"))
				{
					int maxLength = Integer.parseInt(attrs.getValue("size"));
					int strength  = 1;
					if(strength < 1) strength = 1;
					String func = "''";
					int typeValue = getTypeValue(type);
					typeValue = typeValue > -1 ? typeValue : this.fieldIndex;
					BytesFunction bytesFunc = (BytesFunction)compiler.compile(func);
					this.fields.add(new Bytes(typeValue, name, this.fieldIndex, maxLength, strength, false, bytesFunc));
				}
			}
			else if(raw.equals("loader"))
			{
				String src = "true";
				Object o = compiler.compile(src);
				if(o instanceof BooleanFunction)
				{
					this.loadFilter = (BooleanFunction)o;
				}
				else
				{
					throw new SystemException("Gate must be a BOOLEAN_FUNCTION:"+src);
				}

				int mint = parseInt(attrs.getValue("meter"), 1000000);
				this.loadMeter = new Meter(mint);
			}
			else if(raw.equals("load"))
			{
				if(command instanceof LoadCommand)
				{

					String fname = attrs.getValue("name");


					Field field = this.dataRecord.getField(fname);

					if(field == null)
					{
						throw new SystemException("The field '"+fname+"' does not exist");
					}


					if(field instanceof TimeInt)
					{
						TimeInt ifield = (TimeInt)field;
						String src = attrs.getValue("value");


						if(src == null)
						{
							throw new SystemException("time field '"+ifield.getName()+"' must be assigned a time value");
						}

						Object o = compiler.compile(src);
						if(!(o instanceof BytesFunction))
						{
							throw new SystemException("Load macros for Time fields must return bytes.");
						}

						IntFunction func = (IntFunction)o;
						ifield.setLoadFunction(func);
					}
					else if(field instanceof Int)
					{
						Int ifield = (Int)field;
						String src = attrs.getValue("value");
						if(src == null)
						{
							throw new SystemException("int field '"+ifield.getName()+"' must be assigned an int value");
						}

						Object o = compiler.compile(src);

						if(ifield.getDictionary() == null)
						{
							if(!(o instanceof IntFunction))
							{
								throw new SystemException("Load macros for int fields must return int.");
							}
						}
						else
						{
							if(!(o instanceof BytesFunction))
							{
								throw new SystemException("Load macros for int fields with dictionaries must return bytes.");
							}
						}

						IntFunction func = (IntFunction)o;
						ifield.setLoadFunction(func);
					}
					else if(field instanceof Bytes)
					{
						Bytes bfield = (Bytes)field;
						String src = attrs.getValue("value");

						if(src == null)
						{
							throw new SystemException("bytes field '"+bfield.getName()+"' must be assigned a bytes value");
						}


						Object o = compiler.compile(src);
						if(!(o instanceof BytesFunction))
						{
							throw new SystemException("Macros for bytes fields must return bytes.");
						}

						BytesFunction func = (BytesFunction)o;
						bfield.setLoadFunction(func);
					}
					else if(field instanceof LongInt)
					{
						LongInt lfield = (LongInt)field;
						String src = attrs.getValue("value");

						if(src.equals("-"))
						{
							throw new SystemException("long field '"+lfield.getName()+"' must be assigned a long value");
						}

						Object o = compiler.compile(src);
						if(!(o instanceof LongFunction))
						{
							throw new SystemException("Macros for long fields must return long.");
						}

						LongFunction func = (LongFunction)o;
						lfield.setLoadFunction(func);
					}
				}
			}
		}
		catch(SystemException se)
		{
			if(showLocation)
			{
				System.out.println("Configuration error:");
				logLocation();
			}

			System.out.println(se.getMessage());
			Throwable c = se.getCause();
			int i = 1;
			while(c != null)
			{
				System.out.println("-Cause:");
				System.out.println(c.getMessage());
				c = c.getCause();
				i++;
			}
			//se.printStackTrace();
			throw new Error(se.getMessage(), se);
		}
		catch(Throwable e)
		{
			System.out.println("Configuration error:");
			logLocation();
			System.out.println("");
			System.out.println(e.getMessage());

			throw new Error("Error during configuration.", e);
		}
	}

	private void logLocation()
	{
		if(locator != null)
		{
			System.out.println(locator.getSystemId());
			System.out.println("Line number:"+locator.getLineNumber());
		}
	}

	public void endElement(String uri, String local, String raw)
	{
		if(raw.equals("columns"))
		{
			this.dataRecord = new DataRecord(this.fields);
		}
		else if(raw.equals("loader"))
		{
			this.delimitedRecord.setActiveFields((DelimitedField[])this.activeDelimitedFields.toArray(new DelimitedField[activeDelimitedFields.size()]));
		}
		else if(raw.equals("reader"))
		{
			if(command instanceof ReadCommand)
			{
				this.dataRecord.setActiveFields((Field[])this.activeFields.toArray(new Field[activeFields.size()]));
				Reader.dataRecord = this.dataRecord;
				Col.inContext = false;
			}
		}
	}

	public void	error(SAXParseException e)
	{
		System.out.println("SAX Parsing Error");
		System.out.println(e.getMessage());
		if(locator != null)
		{
			System.out.println(locator.getSystemId());
			System.out.println("Line number:"+locator.getLineNumber());
		}
		throw new Error("SAX Eception", e);
	}

	public void	fatalError(SAXParseException e)
	{
		System.out.println("SAX Parsing Error");
		System.out.println(e.getMessage());
		if(locator != null)
		{
			System.out.println(locator.getSystemId());
			System.out.println("Line number:"+locator.getLineNumber());
		}
	}

	public TimeSeries getTimeSeries(String increment)
	{
		int minYYYYMMDD = 200000000;
		int maxYYYYMMDD = 0;

		File file = new File(indexDir);
		File[] years = file.listFiles();
		for(int i=0; i<years.length; i++)
		{
			File year = years[i];
			int YYYY = Integer.parseInt(year.getName());

			File[] months = year.listFiles();
			for(int m=0; m<months.length; m++)
			{
				File month = months[m];

				int MM = Integer.parseInt(month.getName());

				File[] days = month.listFiles();
				for(int d=0; d<days.length; d++)
				{
					File day = days[d];
					int DD = Integer.parseInt(day.getName());
					int YYYYMMDD = (YYYY*10000)+(MM*100)+DD;

					if(YYYYMMDD < minYYYYMMDD)
					{
						minYYYYMMDD = YYYYMMDD;
					}

					if(YYYYMMDD > maxYYYYMMDD)
					{
						maxYYYYMMDD = YYYYMMDD;
					}
				}
			}
		}

		if(increment.equals("days"))
		{
			DaySeries daySeries = new DaySeries();
			return daySeries.getTimeSeries(minYYYYMMDD, maxYYYYMMDD);
		}
		else if(increment.equals("weeks"))
		{

			DaySeries daySeries = new DaySeries();

			int minYYYYWW = daySeries.convertToWeeks(minYYYYMMDD);
			int maxYYYYWW =  daySeries.convertToWeeks(maxYYYYMMDD);
			WeekSeries weekSeries = new WeekSeries();
			return weekSeries.getTimeSeries(minYYYYWW, maxYYYYWW);
		}
		else if(increment.equals("months"))
		{
			DaySeries daySeries = new DaySeries();
			int minYYYYMM = daySeries.convertToMonths(minYYYYMMDD);
			int maxYYYYMM =  daySeries.convertToMonths(maxYYYYMMDD);
			MonthSeries monthSeries = new MonthSeries();
			return monthSeries.getTimeSeries(minYYYYMM, maxYYYYMM);
		}
		else if(increment.equals("quarters"))
		{
			DaySeries daySeries = new DaySeries();

			int minYYYYQQ = daySeries.convertToQuarters(minYYYYMMDD);
			int maxYYYYQQ =  daySeries.convertToQuarters(maxYYYYMMDD);
			QuarterSeries quarterSeries = new QuarterSeries();
			return quarterSeries.getTimeSeries(minYYYYQQ, maxYYYYQQ);
		}
		else if(increment.equals("years"))
		{
			DaySeries daySeries = new DaySeries();

			int minYYYY = daySeries.convertToYears(minYYYYMMDD);
			int maxYYYY =  daySeries.convertToYears(maxYYYYMMDD);

			YearSeries yearSeries = new YearSeries();
			return yearSeries.getTimeSeries(minYYYY, maxYYYY);
		}

		return null;
	}

	private Field getIntField(int size, String type, String name, int fieldIndex, String funcString)
		throws Exception
	{
		IntFunction intFunc = (IntFunction)compiler.compile(funcString);
		int typeValue = getTypeValue(type);

		switch(size)
		{
			case 1:
				return new Int1(typeValue, name, fieldIndex, intFunc);
			case 2:
				return new Int2(typeValue, name, fieldIndex, intFunc);
			case 3:
				return new Int3(typeValue, name, fieldIndex, intFunc);
			case 4:
				return new Int4(typeValue, name, fieldIndex, intFunc);
		}

		return null;
	}

	private int parseInt(String i, int def)
		throws SystemException
	{
		if(i == null)
		{
			return def;
		}
		else
		{
			try
			{
				return Integer.parseInt(i);
			}
			catch(Exception e)
			{
				throw new SystemException("Error parsing int:"+i);
			}
		}
	}



	public int getTypeValue(String typeName)
		throws IOException
	{
		return -1;
	}
}