/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.system;

import amberstone.data.*;
import amberstone.data.fields.*;

import amberstone.lang.*;
import amberstone.read.*;
import amberstone.load.*;
import amberstone.dictionary.*;
import amberstone.summary.*;
import amberstone.transaction.*;
import amberstone.time.*;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.ErrorHandler;
import org.xml.sax.Locator;
import org.xml.sax.helpers.DefaultHandler;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import java.util.LinkedList;
import java.util.ArrayList;
import java.util.Properties;
import java.io.IOException;
import java.io.File;

public class SummarizersConfig extends DefaultHandler implements ErrorHandler
{
	private Config parent;

	private String summariesDir;
	private String indexDir;
	public static String index;
	private Command command;
	private Properties system;
	private String AMBERSTONE_HOME;
	private amberstone.read.Selector select;
	private Set activeFields = new TreeSet();

	private boolean showLocation = true;

	private SummaryThread currentSummaryThread;
	private List<SummaryThread> summaryThreads = new ArrayList();
	private List<FileTemplate> topFiles = new ArrayList();
	private List<PropTemplate> tprops = new ArrayList();
	private FileTemplate currentFileTemplate;
	private TransactionTrackerTemplate currentTransaction;
	private Meter readMeter;
	private BooleanFunction threadFilter;
	private Map<String, String> summaries = new HashMap();
	private int threadNum;
	private amberstone.lang.Compiler compiler;
	private DataRecord dataRecord;
	private Locator locator;
	private LongFunction grain;

	public SummarizersConfig(Config parent,
					         SummarizeCommand command,
	                         amberstone.lang.Compiler compiler,
	                         Properties system)
		throws Exception
	{
		this.AMBERSTONE_HOME = AmberstoneConfig.AMBERSTONE_HOME;
		this.parent = parent;
		this.command = command;
		this.compiler = compiler;
		this.system = system;

		this.dataRecord = parent.getIndexConfig().getDataRecord();
		this.index = command.getIndex();
		String configFile = command.getConfig();
		this.indexDir = AMBERSTONE_HOME + File.separatorChar + "storage" + File.separatorChar + index + File.separatorChar + "index";
		String config = AMBERSTONE_HOME + File.separatorChar + "aggregates" + File.separatorChar + "conf" + File.separatorChar + configFile;
		this.summariesDir = AMBERSTONE_HOME + File.separatorChar + "aggregates" + File.separatorChar + "in";
		parse(config);
	}

	public String getIndex()
	{
		return this.index;
	}

	public String getIndexDir()
	{
		return this.indexDir;
	}

	public String getSummariesDir()
	{
		return this.summariesDir;
	}

	public SummaryThread[] getSummaryThreads()
	{
		return (SummaryThread[])this.summaryThreads.toArray(new SummaryThread[this.summaryThreads.size()]);
	}

	public void parse(String path) throws Exception
	{
		XMLReader parser = (XMLReader)Class.forName("org.apache.xerces.parsers.SAXParser").newInstance();
	  	parser.setFeature("http://xml.org/sax/features/validation", true);
		parser.setContentHandler(this);
		parser.setErrorHandler(this);
		parser.parse(path);
	}

	public void setDocumentLocator(Locator locator)
	{
		this.locator = locator;
	}

	public DataRecord getDataRecord()
	{
		return this.dataRecord;
	}

	public void startElement(String uri,
						     String local,
							 String raw,
	                         Attributes attrs)
	{
		try
		{
			if(raw.equals("aggregators"))
			{
				Col.inContext = true;
			}
			else if(raw.equals("aggregator"))
			{
				this.dataRecord = this.dataRecord.cloneRecord();
				this.activeFields = new TreeSet();
				this.topFiles = new ArrayList();
				Col.setDataRecord(this.dataRecord);
				Col.setActiveFields(this.activeFields);

				String src = attrs.getValue("grain");
				Object o = compiler.compile(src);

				if(!(o instanceof LongFunction))
				{
					throw new SystemException("Invalid function type, LONG_FUNCTION expected: "+src);
				}

				this.grain = (LongFunction)o;
			}
			else if(raw.equals("cube"))
			{
				String name = attrs.getValue("name");
				summaries.put(name, name);

				int maxSize = Integer.parseInt(attrs.getValue("maxSize"));

				String src = attrs.getValue("aggregate");
				Object o = compiler.compile(src);

				if(!(o instanceof SummaryFunction))
				{
					throw new SystemException("Invalid function type, SUMMARY_FUNCTION expected: "+src);
				}

				SummaryFunction func = (SummaryFunction)o;

				CubeTemplate ht = new CubeTemplate(name, maxSize);

				src = attrs.getValue("where");
				o = compiler.compile(src);

				if(!(o instanceof BooleanFunction))
				{
					throw new SystemException("Invalid function type, BOOLEAN_FUNCTION expected: "+src);
				}

				BooleanFunction filter = (BooleanFunction)o;

				src = attrs.getValue("groupby");
				o = compiler.compile(src);

				if(!(o instanceof KeyFunction))
				{
					throw new SystemException("Invalid function type, KEY_FUNCTION expected: "+src);
				}

				KeyFunction keyFunc = (KeyFunction)o;

				ht.setOperator(func);
				ht.setFilter(filter);
				ht.setKeyFunction(keyFunc);

				if(this.currentTransaction == null)
				{
					topFiles.add(ht);
				}
				else
				{
					this.currentTransaction.addFileTemplate(ht);
				}
			}
			else if(raw.equals("transaction"))
			{
				Object o  = compiler.compile(attrs.getValue("key"));
				KeyFunction id = null;
				if(o instanceof KeyFunction)
				{
					id = (KeyFunction)o;
				}
				else
				{
					throw new SystemException("Transaction key must be a KEY_FUNCTION:"+attrs.getValue("key"));
				}

				o = compiler.compile(attrs.getValue("time"));

				TimeFunction time = null;
				if(o instanceof TimeFunction)
				{
					time = (TimeFunction)o;
				}
				else
				{
					throw new SystemException("Transaction time must be an TIME_FUNCTION:"+attrs.getValue("time"));
				}

				int expires      = parseInt(attrs.getValue("timeout"), -1);
				int size         = parseInt(attrs.getValue("size"), -1);

				TransactionTrackerTemplate transactionTrackerTemplate = new TransactionTrackerTemplate(id, time, expires, size);
				this.currentTransaction = transactionTrackerTemplate;
				TransactionFunction.setTracker(this.currentTransaction);
				Col.inContext = false;
			}
			else if(raw.equals("property"))
			{

				SummarizeCommand summarizeCommand = (SummarizeCommand)command;

				Col.inContext = true;

				String name         = attrs.getValue("name");
				String funcString   = attrs.getValue("value");

				Object o = compiler.compile(funcString);
				Prop prop = null;
				if(o instanceof Prop)
				{
					prop = (Prop)o;
				}
				else
				{
					throw new SystemException("Transaction prop must be a PROP_FUNCTION:"+funcString);
				}

				PropTemplate p = new PropTemplate(new StringLiteral(name).getKey(), prop);
				tprops.add(p);
				Col.inContext = false;
			}
		}
		catch(SystemException se)
		{
			if(showLocation)
			{
				System.out.println("Configuration error:");
				logLocation();
			}

			System.out.println(se.getMessage());
			Throwable c = se.getCause();
			int i = 1;
			while(c != null)
			{
				System.out.println("-Cause:");
				System.out.println(c.getMessage());
				c = c.getCause();
				i++;
			}
			//se.printStackTrace();
			throw new Error(se.getMessage(), se);
		}
		catch(Throwable e)
		{
			System.out.println("Configuration error:");
			logLocation();
			System.out.println("");
			System.out.println(e.getMessage());

			throw new Error("Error during configuration.", e);
		}
	}

	private void logLocation()
	{
		if(locator != null)
		{
			System.out.println(locator.getSystemId());
			System.out.println("Line number:"+locator.getLineNumber());
		}
	}

	public void endElement(String uri, String local, String raw)
	{

		if(raw.equals("transaction"))
		{
			if(this.currentTransaction.fileTemplateCount() > 0)
			{
				topFiles.add(this.currentTransaction);
			}

			tprops.clear();
			this.currentTransaction = null;
			TransactionFunction.setTracker(null);
			Col.inContext = true;
		}
		else if(raw.equals("properties"))
		{
			PropTemplate[] tp = (PropTemplate[])this.tprops.toArray(new PropTemplate[tprops.size()]);
			this.currentTransaction.setProps(tp);
		}
		else if(raw.equals("aggregator"))
		{
			FileTemplate[] topTemplates = (FileTemplate[])this.topFiles.toArray(new FileTemplate[this.topFiles.size()]);
			this.dataRecord.setActiveFields((Field[])this.activeFields.toArray(new Field[activeFields.size()]));
			SummaryThread summaryThread = new SummaryThread(parent, this.dataRecord, topTemplates);
			summaryThread.setGrain(this.grain);
			summaryThreads.add(summaryThread);
			activeFields.clear();
			++threadNum;

		}
		else if(raw.equals("aggregators"))
		{
			Col.inContext = false;
		}
	}

	public void	error(SAXParseException e)
	{
		System.out.println("SAX Parsing Error");
		System.out.println(e.getMessage());
		if(locator != null)
		{
			System.out.println(locator.getSystemId());
			System.out.println("Line number:"+locator.getLineNumber());
		}
		throw new Error("SAX Eception", e);
	}

	public void	fatalError(SAXParseException e)
	{
		System.out.println("SAX Parsing Error");
		System.out.println(e.getMessage());
		if(locator != null)
		{
			System.out.println(locator.getSystemId());
			System.out.println("Line number:"+locator.getLineNumber());
		}
	}

	private int parseInt(String i, int def)
		throws SystemException
	{
		if(i == null)
		{
			return def;
		}
		else
		{
			try
			{
				return Integer.parseInt(i);
			}
			catch(Exception e)
			{
				throw new SystemException("Error parsing int:"+i);
			}
		}
	}

	public int getTypeValue(String typeName)
		throws IOException
	{
		return -1;
	}
}