/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.time;

import amberstone.lang.*;
import amberstone.report.*;

import java.util.Arrays;
import java.util.ArrayList;
import java.util.List;
import java.util.List;
import java.util.Calendar;

public class DaySeries extends TimeSeries
{
	private static final FunctionType[][] allow = {{FunctionType.STRING_LITERAL, FunctionType.STRING_LITERAL},
												   {FunctionType.DAY, FunctionType.DAY},
												   {FunctionType.STRING_LITERAL, FunctionType.DAY},
												   {FunctionType.DAY, FunctionType.STRING_LITERAL}};

	public FunctionType[][] strict()
	{
		return allow;
	}

	public String output(String YYYYMMDD)
	{
		String out = YYYYMMDD.substring(4,6)+"-"+YYYYMMDD.substring(6,8)+"-"+YYYYMMDD.substring(0, 4);
		return out;
	}

	public DaySeries(Increment increment)
	{
		this.increment   = increment;
		start            = increment.getIncrement();
		end              = increment.getIncrement();
	}

	public DaySeries()
	{

	}

	public int granularity()
	{
		return 1;
	}

	public TimeSeries getTimeSeries(int start, int end)
	{
		DaySeries daySeries = new DaySeries();
		daySeries.start = start;
		daySeries.end = end;
		return daySeries;
	}

	public void params(Function[] params) throws ParameterException
	{
		try
		{
			start = validateAndFormat(params[0].toString());
		}
		catch(Exception e)
		{
			throw new ParameterException(params[0], "Invalid day format (YYYY-MM-DD): "+params[0].toString(), e);
		}

		try
		{
			end = validateAndFormat(params[1].toString());
		}
		catch(Exception e)
		{
			throw new ParameterException(params[1], "Invalid day format (YYYY-MM-DD): "+params[1].toString(), e);
		}

		if(start > end)
		{
			throw new ParameterException(params[0], "Start day must be less then or equal to end day");
		}
	}

	public static int validateAndFormat(String time)
		throws Exception
	{
		String[] parts = time.split("-");

		if(parts.length != 3)
		{
			throw new Exception("Day must include three sections");
		}

		int YYYY = validateAndFormatYear(parts[0]);
		int MM   = validateAndFormatMonth(parts[1]);
		int DD   = validateAndFormatDay(parts[2]);

		return (YYYY*10000)+(MM*100)+DD;
	}

	public String toString()
	{
		return "days";
	}

	public int convertTo(TimeSeries convertSeries, int convertValue)
	{
		return convertSeries.convertToDays(convertValue);
	}

	public int convertToDays(int YYYYMMDD)
	{
		return YYYYMMDD;
	}

	public int convertToMonths(int YYYYMMDD)
	{
		int YYYY = (int)(YYYYMMDD/10000);
		int MM = (int)((YYYYMMDD % (YYYY*10000)) / 100);

		return (YYYY*100)+MM;
	}

	public int getTime(int YYYYMMDD)
	{
		int YYYY = (int)(YYYYMMDD/10000);
		int MM = (int)((YYYYMMDD % (YYYY*10000)) / 100);
		int DD = YYYYMMDD%(MM*100);
		cal.clear();
		cal.set(YYYY, MM-1, DD, 15, 0, 0);
		return (int)cal.getTimeInMillis()/1000;
	}

	public int convertToQuarters(int YYYYMMDD)
	{
		int YYYY = (int)(YYYYMMDD/10000);
		int MM = (int)((YYYYMMDD % (YYYY*10000)) / 100);

		switch(MM)
		{
			case 1:
				return (YYYY*10)+1;
			case 2:
				return (YYYY*10)+1;
			case 3:
				return (YYYY*10)+1;
			case 4:
				return (YYYY*10)+2;
			case 5:
				return (YYYY*10)+2;
			case 6:
				return (YYYY*10)+2;
			case 7:
				return (YYYY*10)+3;
			case 8:
				return (YYYY*10)+3;
			case 9:
				return (YYYY*10)+3;
			case 10:
				return (YYYY*10)+4;
			case 11:
				return (YYYY*10)+4;
			case 12:
				return (YYYY*10)+4;
		}

		return -1;
	}

	public int getQuarter(int MM)
	{
		switch(MM)
		{
			case 1:
				return 1;
			case 2:
				return 1;
			case 3:
				return 1;
			case 4:
				return 2;
			case 5:
				return 2;
			case 6:
				return 2;
			case 7:
				return 3;
			case 8:
				return 3;
			case 9:
				return 3;
			case 10:
				return 4;
			case 11:
				return 4;
			case 12:
				return 4;
		}

		return -1;
	}

	public int getWeek(int YYYYMMDD)
	{
		int YYYY = (int)(YYYYMMDD/10000);
		int MM = (int)((YYYYMMDD % (YYYY*10000)) / 100);
		int DD = YYYYMMDD % ((YYYY*10000)+(MM*100));

		cal.set(YYYY, MM, DD);
		int WW = cal.get(cal.WEEK_OF_YEAR);
		return WW;
	}

	public int getDayOfWeek(int YYYYMMDD)
	{
		int YYYY = (int)(YYYYMMDD/10000);
		int MM = (int)((YYYYMMDD % (YYYY*10000)) / 100);
		int DD = YYYYMMDD % ((YYYY*10000)+(MM*100));

		cal.set(YYYY, MM, DD);
		int d = cal.get(cal.DAY_OF_WEEK);
		return d;
	}

	public int convertToWeeks(int YYYYMMDD)
	{
		int YYYY = (int)(YYYYMMDD/10000);
		int MM = (int)((YYYYMMDD % (YYYY*10000)) / 100);
		int DD = YYYYMMDD % ((YYYY*10000)+(MM*100));

		cal.set(YYYY, MM, DD);
		int WW = cal.get(cal.WEEK_OF_YEAR);
		return (YYYY*100)+WW;
	}

	public int currentTime()
	{
		cal.setTimeInMillis(System.currentTimeMillis());
		int year  = cal.get(cal.YEAR);
		int month = cal.get(cal.MONTH)+1;
		int day = cal.get(cal.DAY_OF_MONTH);
		return (year*10000)+(month*100)+day;
	}


	public int convertToYears(int YYYYMMDD)
	{
		return (int)(YYYYMMDD/10000);
	}

	public int convertYYYYMMDD(int YYYYMMDD)
	{
		return YYYYMMDD;
	}

	public int getMaxYYYYMMDD()
	{
		return end;
	}

	public int getMinYYYYMMDD()
	{
		return start;
	}

	public void setRolling(boolean rolling)
	{
		if(rolling)
		{
			cal.setTimeInMillis(System.currentTimeMillis());
			int YYYY = cal.get(cal.YEAR);
			int MM = cal.get(cal.MONTH);
			int DD =  cal.get(cal.DAY_OF_MONTH);
			this.end = (YYYY*10000)+((MM+1)*100)+DD;
		}
	}

	public int[] getCols(int minYYYYMMDD, int maxYYYYMMDD)
	{
		int minYYYY = (int)(minYYYYMMDD/10000);
		int minMM = (int)((minYYYYMMDD % (minYYYY*10000)) / 100);
		int minDD = minYYYYMMDD % ((minYYYY*10000)+(minMM*100));

		ArrayList<Integer> days = new ArrayList();

		cal.set(minYYYY, minMM-1, minDD);
		while(true)
		{
			int testYYYY = cal.get(Calendar.YEAR);
			int testMM = cal.get(Calendar.MONTH);
			int testDD = cal.get(Calendar.DAY_OF_MONTH);

			int testYYYYMMDD = (testYYYY*10000)+((testMM+1)*100)+testDD;

			if(testYYYYMMDD <= maxYYYYMMDD)
			{
				days.add(new Integer(testYYYYMMDD));
				cal.add(Calendar.DATE, 1);
			}
			else
			{
				break;
			}
		}

		int[] d = new int[days.size()];
		for(int i=0; i<d.length; i++)
		{
			d[i] = days.get(i).intValue();
		}

		return d;
	}
}