/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.time;

import amberstone.lang.*;
import amberstone.report.*;

import java.util.Calendar;
import java.util.Arrays;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;

public class MonthSeries extends TimeSeries
{

	private boolean incremental;
	private static final int[] lastDayOfMonth = {30, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

	private static final FunctionType[][] allow = {{FunctionType.STRING_LITERAL, FunctionType.STRING_LITERAL},
												   {FunctionType.MONTH, FunctionType.MONTH},
												   {FunctionType.STRING_LITERAL, FunctionType.MONTH},
												   {FunctionType.MONTH, FunctionType.STRING_LITERAL}};

	public FunctionType[][] strict()
	{
		return allow;
	}

	public boolean incremental()
	{
		return this.incremental;
	}

	public MonthSeries()
	{

	}

	public String output(String YYYYWW)
	{
		String out = YYYYWW.substring(5,7)+"-"+YYYYWW.substring(0, 4);
		return out;
	}

	public int granularity()
	{
		return 3;
	}

	public TimeSeries getTimeSeries(int start, int end)
	{
		MonthSeries monthSeries = new MonthSeries();
		monthSeries.start = start;
		monthSeries.end = end;
		return monthSeries;
	}

	public MonthSeries(Month month)
	{
		increment = month;
		start = end = month.getIncrement();
	}

	public void params(Function[] params) throws ParameterException
	{
		try
		{
			start = validateAndFormat(params[0].toString());
		}
		catch(Exception e)
		{
			throw new ParameterException(params[0], "Invalid month format (YYYY-MM): "+params[0].toString(), e);
		}

		try
		{
			end = validateAndFormat(params[1].toString());
		}
		catch(Exception e)
		{
			throw new ParameterException(params[1], "Invalid month format (YYYY-MM): "+params[1].toString(), e);
		}

		if(start > end)
		{
			throw new ParameterException(params[0], "Start month must be less then or equal to end month");
		}

		if(params[1] instanceof Month)
		{
			this.incremental = true;
		}
	}

	public static int validateAndFormat(String time)
		throws Exception
	{
		String[] parts = time.split("-");

		if(parts.length != 2)
		{
			throw new Exception("Month must include two sections");
		}

		int YYYY = validateAndFormatYear(parts[0]);
		int MM = validateAndFormatMonth(parts[1]);

		return (YYYY*100)+MM;
	}

	public String toString()
	{
		return "months";
	}

	public int convertYYYYMMDD(int YYYYMMDD)
	{
		return (int)YYYYMMDD/100;
	}

	public int convertTo(TimeSeries convertSeries, int convertValue)
	{
		return convertSeries.convertToMonths(convertValue);
	}

	public int convertToMonths(int YYYYMM)
	{
		return YYYYMM;
	}

	public int getTime(int YYYYMM)
	{
		int YYYY = (int)(YYYYMM/100);
		int MM = (int)(YYYYMM % (YYYY*100));
		int DD = 15;
		cal.clear();
		cal.set(YYYY, MM-1, DD, 15, 0, 0);
		return (int)cal.getTimeInMillis()/1000;
	}


	public int convertToYears(int YYYYMM)
	{
		return (int)(YYYYMM/100);
	}

	public int convertToQuarters(int YYYYMM)
	{
		int YYYY = (int)(YYYYMM/100);
		int MM = YYYYMM % (YYYY*100);
		switch(MM)
		{
			case 1:
				return (YYYY*10)+1;
			case 2:
				return (YYYY*10)+1;
			case 3:
				return (YYYY*10)+1;
			case 4:
				return (YYYY*10)+2;
			case 5:
				return (YYYY*10)+2;
			case 6:
				return (YYYY*10)+2;
			case 7:
				return (YYYY*10)+3;
			case 8:
				return (YYYY*10)+3;
			case 9:
				return (YYYY*10)+3;
			case 10:
				return (YYYY*10)+4;
			case 11:
				return (YYYY*10)+4;
			case 12:
				return (YYYY*10)+4;
		}

		return -1;
	}

	public void setRolling(boolean rolling)
	{
		if(rolling)
		{
			cal.setTimeInMillis(System.currentTimeMillis());
			int YYYY = cal.get(cal.YEAR);
			int MM = cal.get(cal.MONTH);
			this.end = (YYYY*100)+(MM+1);
		}
	}

	public int[] getCols(int minYYYYMMDD, int maxYYYYMMDD)
	{
		int minYYYY = (int)(minYYYYMMDD/10000);
		int minMM = (int)((minYYYYMMDD % (minYYYY*10000)) / 100);
		int minDD = minYYYYMMDD % ((minYYYY*10000)+(minMM*100));

		Set<Integer> months = new HashSet();

		cal.set(minYYYY, minMM-1, minDD);
		while(true)
		{
			int testYYYY = cal.get(Calendar.YEAR);
			int testMM = cal.get(Calendar.MONTH);
			int testDD = cal.get(Calendar.DAY_OF_MONTH);

			int testYYYYMMDD = (testYYYY*10000)+((testMM+1)*100)+testDD;

			if(testYYYYMMDD <= maxYYYYMMDD)
			{
				months.add(new Integer((testYYYY*100)+(testMM+1)));
				cal.add(Calendar.DATE, 1);
			}
			else
			{
				break;
			}
		}

		int[] m = new int[months.size()];

		Iterator<Integer> it = months.iterator();
		int i = -1;
		while(it.hasNext())
		{
			m[++i] = it.next().intValue();
		}

		Arrays.sort(m);

		return m;
	}

	public int currentTime()
	{
		cal.setTimeInMillis(System.currentTimeMillis());
		int year  = cal.get(cal.YEAR);
		int month = cal.get(cal.MONTH)+1;
		return (year*100)+month;
	}


	public int getMinYYYYMMDD()
	{
		return (start*100)+1;
	}

	public int getMaxYYYYMMDD()
	{
		int YYYY = (int)end/100;
		int MM = end%(YYYY*100);
		//System.out.println("MONTH:"+end);
		return (end*100) + lastDayOfMonth[(MM)-1];
	}
}