/*
*
*	Copyright 2012 Joel Bernstein
*
*	Licensed under the Apache License, Version 2.0 (the "License");
*	you may not use this file except in compliance with the License.
*	You may obtain a copy of the License at
*
*	 http://www.apache.org/licenses/LICENSE-2.0
*
*	Unless required by applicable law or agreed to in writing, software
*	distributed under the License is distributed on an "AS IS" BASIS,
*	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*	See the License for the specific language governing permissions and
*	limitations under the License.
*
*/

package amberstone.time;

import amberstone.lang.*;
import amberstone.report.*;

import java.util.Calendar;
import java.util.Arrays;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;
import java.util.TreeSet;
import java.util.Iterator;

public class QuarterSeries extends TimeSeries
{

	private boolean incremental;

	private static final FunctionType[][] allow = {{FunctionType.STRING_LITERAL, FunctionType.STRING_LITERAL},
												   {FunctionType.QUARTER, FunctionType.QUARTER},
												   {FunctionType.STRING_LITERAL, FunctionType.QUARTER},
												   {FunctionType.QUARTER, FunctionType.STRING_LITERAL}};

	public FunctionType[][] strict()
	{
		return allow;
	}

	public QuarterSeries()
	{

	}

	public String output(String YYYYQQ)
	{
		String out = YYYYQQ.substring(5,7)+"-"+YYYYQQ.substring(0, 4);
		return out;
	}

	public int granularity()
	{
		return 4;
	}

	public TimeSeries getTimeSeries(int start, int end)
	{
		QuarterSeries quarterSeries = new QuarterSeries();
		quarterSeries.start = start;
		quarterSeries.end = end;
		return quarterSeries;
	}


	public QuarterSeries(Quarter quarter)
	{
		increment = quarter;
		start = end = quarter.getIncrement();
	}


	public void params(Function[] params) throws ParameterException
	{
		try
		{
			start = validateAndFormat(params[0].toString());
		}
		catch(Exception e)
		{
			throw new ParameterException(params[0], "Invalid quarter format (YYYY-Q): "+params[0].toString(), e);
		}

		try
		{
			end = validateAndFormat(params[1].toString());
		}
		catch(Exception e)
		{
			throw new ParameterException(params[1], "Invalid quarter format (YYYY-Q): "+params[1].toString(), e);
		}

		if(start > end)
		{
			throw new ParameterException(params[0], "Start quarter must be less then or equal to end quarter");
		}

		if(params[1] instanceof Quarter)
		{
			this.incremental = true;
		}
	}

	public static int validateAndFormat(String time)
		throws Exception
	{
		String[] parts = time.split("-");

		if(parts.length != 2)
		{
			throw new Exception("Quarter must include two sections");
		}

		int YYYY = validateAndFormatYear(parts[0]);
		int Q = validateAndFormatQuarter(parts[1]);
		return (YYYY*10)+Q;
	}

	public String toString()
	{
		return "quarters";
	}

	public void setRolling(boolean rolling)
	{
		if(rolling)
		{
			cal.setTimeInMillis(System.currentTimeMillis());
			int YYYY = cal.get(cal.YEAR);
			int MM = cal.get(cal.MONTH);
			this.end = getQuarterFromMonth(YYYY, ++MM);
		}
	}

	public int getTime(int YYYYQQ)
	{
		int YYYY = (int)(YYYYQQ/100);
		int QQ = (int)(YYYYQQ % (YYYY*100));
		int MM = month(QQ);
		cal.clear();
		cal.set(YYYY, MM, 15, 15, 0, 0);
		return (int)cal.getTimeInMillis()/1000;
	}


	public int convertTo(TimeSeries convertSeries, int convertValue)
	{
		return convertSeries.convertToQuarters(convertValue);
	}

	public int convertToQuarters(int YYYYQ)
	{
		return YYYYQ;
	}

	public int convertToYears(int YYYYQ)
	{
		return (int)YYYYQ/10;
	}

	public int convertYYYYMMDD(int YYYYMMDD)
	{
		int YYYY = (int)(YYYYMMDD/10000);
		int MM = (int)((YYYYMMDD % (YYYY*10000)) / 100);
		return getQuarterFromMonth(YYYY, MM);
	}

	public int getMinYYYYMMDD()
	{
		int year = (int)(start / 10);
		int quarter = start % (year*10);

		switch(quarter)
		{
			case 1:
				return (year*10000)+101;
			case 2:
				return (year*10000)+401;
			case 3:
				return (year*10000)+701;
			case 4:
				return (year*10000)+1001;
		}

		return -1;
	}

	public int[] getCols(int minYYYYMMDD, int maxYYYYMMDD)
	{
		//System.out.println("Min Quarter:"+minYYYYMMDD+" Max Quarter:"+maxYYYYMMDD);
		int minYYYY = (int)(minYYYYMMDD/10000);
		int minMM = (int)((minYYYYMMDD % (minYYYY*10000)) / 100);
		int minDD = minYYYYMMDD % ((minYYYY*10000)+(minMM*100));

		Set<Integer> quarters = new HashSet();

		cal.set(minYYYY, minMM-1, minDD);
		while(true)
		{
			int testYYYY = cal.get(Calendar.YEAR);
			int testMM = cal.get(Calendar.MONTH);
			int testDD = cal.get(Calendar.DAY_OF_MONTH);

			int testYYYYMMDD = (testYYYY*10000)+((testMM+1)*100)+testDD;

			if(testYYYYMMDD <= maxYYYYMMDD)
			{
				quarters.add(new Integer(getQuarterFromMonth(testYYYY, testMM+1)));
				cal.add(Calendar.DATE, 1);
			}
			else
			{
				break;
			}
		}

		int[] q = new int[quarters.size()];

		Iterator<Integer> it = quarters.iterator();
		int i = -1;
		while(it.hasNext())
		{
			q[++i] = it.next().intValue();
		}

		Arrays.sort(q);

		return q;
	}

	public int getMaxYYYYMMDD()
	{
		int year = (int)(end / 10);
		int quarter = end % (year*10);

		switch(quarter)
		{
			case 1:
				return (year*10000)+331;
			case 2:
				return (year*10000)+0630;
			case 3:
				return (year*10000)+930;
			case 4:
				return (year*10000)+1231;
		}

		return -1;
	}

	public int currentTime()
	{
		cal.setTimeInMillis(System.currentTimeMillis());
		int year  = cal.get(cal.YEAR);
		int month = cal.get(cal.MONTH)+1;
		return getQuarterFromMonth(year, month);
	}

	public static int getQuarterFromMonth(int year, int month)
	{
		int q = -1;

		switch(month)
		{
			case 1:
				q = 1;
				break;
			case 2:
				q = 1;
				break;
			case 3:
				q = 1;
				break;
			case 4:
				q = 2;
				break;
			case 5:
				q = 2;
				break;
			case 6:
				q = 2;
				break;
			case 7:
				q = 3;
				break;
			case 8:
				q = 3;
				break;
			case 9:
				q = 3;
				break;
			case 10:
				q = 4;
				break;
			case 11:
				q = 4;
				break;
			case 12:
				q = 4;
				break;
		}

		return (year*10)+q;
	}

	private int month(int quarter)
	{
		int month = -1;

		switch(quarter)
		{
			case 1:
				return 0;
			case 2:
				return 3;
			case 3:
				 return 6;
			case 4:
				 return 9;
		}

		return -1;
	}
}